from ._core.grid_ve import GridVolumetricEnergy

from ._ui.param_handler import ParamHandlerVeins
from ._ui.app import AppVeins


############################# CONFIG FILE GLOBALS ##############################
__config_keys__ = set()


######################## COMMAND LINE ARGUMENTS GLOBALS ########################
### These are global variables that are to be set by
### an instance of ParamHandler (or its inherited classes)

MODE: str = '' # mode of the application, i.e. "energies"

import pathlib as _pathlib
PATH_STRUCT:       _pathlib.Path = None # "path/input/structure.pdb"
PATH_ENERGIES_CSV: _pathlib.Path = None # "path/input/energies.csv"
PATH_TRAJ:         _pathlib.Path = None # "path/input/traj.xtc"
FOLDER_OUT:        _pathlib.Path = None # "path/output/"

ENERGY_CUTOFF: float # Energies below this cutoff will be ignored (default 1e-3)



import volgrids as vg

# //////////////////////////////////////////////////////////////////////////////
class ParserConfig(vg.ParserIni):
    def apply_config(self,
        scope_module: dict[str, ],
        scope_dependencies: dict[str, ],
        this_module_keys: set[str],
        all_known_keys: set[str],
    ) -> None:
        """
        Applies the configuration to the provided global dictionary.
        """
        for section in self._ini_sections.keys():
            for k, value in self.iter_splitted_lines(section):
                k = k.upper()
                if k not in this_module_keys:
                    continue
                if k not in all_known_keys:
                    raise ValueError(f"Unknown configuration: {k}.")
                scope_module[k] = self._parse_str(scope_dependencies, value)


    # --------------------------------------------------------------------------
    @staticmethod
    def _parse_str(scope: dict, str_value: str):
        ### INTEGERS
        if str_value.isdigit():
            return int(str_value)

        ### FLOATS
        try: return float(str_value)
        except ValueError: pass

        ### BOOLEANS
        if str_value.lower() in ["true", "false"]:
            return str_value.lower() == "true"

        ### STRINGS
        stripped = str_value.strip('"')
        if len(stripped) != len(str_value):
            return stripped

        stripped = str_value.strip("'")
        if len(stripped) != len(str_value):
            return stripped

        ### MODULE ATTRIBUTES
        error_message = f"Invalid configuration value: {str_value}"

        parts = str_value.split('.')
        if len(parts) < 2: raise ValueError(error_message)

        key = parts.pop(0)
        node = scope.get(key, None)
        if node is None: raise ValueError(error_message)

        while parts:
            key = parts.pop(0)
            try:
                node = node.__dict__.get(key, None)
            except (AttributeError):
                raise ValueError(error_message)
            if node is None:
                raise ValueError(error_message)
        return node


# //////////////////////////////////////////////////////////////////////////////



import numpy as np
from pathlib import Path

import volgrids as vg

# //////////////////////////////////////////////////////////////////////////////
class Grid:
    def __init__(self, ms: "vg.MolSystem", init_grid = True, dtype = None):
        self.ms = ms
        self.xres, self.yres, self.zres = ms.resolution
        self.xmin, self.ymin, self.zmin = ms.minCoords
        self.xmax, self.ymax, self.zmax = ms.maxCoords
        self.dx, self.dy, self.dz = ms.deltas

        if dtype is None: dtype = vg.FLOAT_DTYPE
        self.grid = np.zeros(ms.resolution, dtype = dtype) if init_grid else None
        self.dtype = dtype
        self.fmt: vg.GridFormat = None


    # --------------------------------------------------------------------------
    def __add__(self, other: "Grid|float|int") -> "Grid":
        obj = Grid(self.ms, init_grid = False)
        if isinstance(other, Grid):
            obj.grid = self.grid + other.grid
            return obj
        try:
            obj.grid = self.grid + other
            return obj
        except TypeError:
            raise TypeError(f"Cannot add {type(other)} to Grid. Use another Grid or a numeric value.")


    # --------------------------------------------------------------------------
    def __sub__(self, other: "Grid|float|int") -> "Grid":
        obj = Grid(self.ms, init_grid = False)
        if isinstance(other, Grid):
            obj.grid = self.grid - other.grid
            return obj
        try:
            obj.grid = self.grid - other
            return obj
        except TypeError:
            raise TypeError(f"Cannot substract {type(other)} from Grid. Use another Grid or a numeric value.")


    # --------------------------------------------------------------------------
    def __abs__(self) -> "Grid":
        obj = Grid(self.ms, init_grid = False)
        obj.grid = np.abs(self.grid)
        return obj


    # --------------------------------------------------------------------------
    @classmethod
    def reverse(cls, other: "Grid") -> "Grid":
        """Return a new Grid with the reversed values of the other Grid.
        For boolean grids, the reverse is the logical not.
        For numeric grids, the reverse is the negation of the values.
        """
        obj = cls(other.ms, init_grid = False)
        obj.grid = np.logical_not(other.grid) if (other.dtype == bool) else -other.grid
        return obj


    # --------------------------------------------------------------------------
    def copy(self):
        obj = Grid(self.ms, init_grid = False)
        obj.grid = np.copy(self.grid)
        return obj


    # --------------------------------------------------------------------------
    def is_empty(self):
        return np.all(self.grid == 0)


    # --------------------------------------------------------------------------
    def get_deltas    (self): return np.array((self.dx  , self.dy  , self.dz  ))
    def get_resolution(self): return np.array((self.xres, self.yres, self.zres))
    def get_min_coords(self): return np.array((self.xmin, self.ymin, self.zmin))
    def get_max_coords(self): return np.array((self.xmax, self.ymax, self.zmax))


    # --------------------------------------------------------------------------
    def reshape(self, new_min: tuple[float], new_max: tuple[float], new_res: tuple[float]):
        new_xmin, new_ymin, new_zmin = new_min
        new_xmax, new_ymax, new_zmax = new_max
        new_xres, new_yres, new_zres = new_res

        self.grid = vg.Math.interpolate_3d(
            x0 = np.linspace(self.xmin, self.xmax, self.xres),
            y0 = np.linspace(self.ymin, self.ymax, self.yres),
            z0 = np.linspace(self.zmin, self.zmax, self.zres),
            data_0 = self.grid,
            new_coords = np.mgrid[
                new_xmin : new_xmax : complex(0, new_xres),
                new_ymin : new_ymax : complex(0, new_yres),
                new_zmin : new_zmax : complex(0, new_zres),
            ].T
        ).astype(vg.FLOAT_DTYPE)

        self.xmin, self.ymin, self.zmin = new_xmin, new_ymin, new_zmin
        self.xmax, self.ymax, self.zmax = new_xmax, new_ymax, new_zmax
        self.xres, self.yres, self.zres = new_xres, new_yres, new_zres
        self.dx = (self.xmax - self.xmin) / (self.xres - 1)
        self.dy = (self.ymax - self.ymin) / (self.yres - 1)
        self.dz = (self.zmax - self.zmin) / (self.zres - 1)

        self.ms.minCoords = np.array([self.xmin, self.ymin, self.zmin])
        self.ms.maxCoords = np.array([self.xmax, self.ymax, self.zmax])
        self.ms.resolution = np.array([self.xres, self.yres, self.zres])
        self.ms.deltas = np.array([self.dx, self.dy, self.dz])


    # --------------------------------------------------------------------------
    def reshape_as_other(self, other: "Grid"):
        self.reshape(
            new_min = (other.xmin, other.ymin, other.zmin),
            new_max = (other.xmax, other.ymax, other.zmax),
            new_res = (other.xres, other.yres, other.zres)
        )


    # --------------------------------------------------------------------------
    def save_data(self, folder_out: Path, title: str):
        path_prefix = folder_out / f"{self.ms.molname}.{title}"

        if self.ms.do_traj:
            ### ignore the OUTPUT flag, CMAP is the only format that supports multiple frames
            vg.GridIO.write_cmap(f"{path_prefix}.cmap", self, f"{self.ms.molname}.{self.ms.frame:04}")
            return

        if vg.OUTPUT_FORMAT == vg.GridFormat.DX:
            vg.GridIO.write_dx(f"{path_prefix}.dx", self)
            return

        if vg.OUTPUT_FORMAT == vg.GridFormat.MRC:
            vg.GridIO.write_mrc(f"{path_prefix}.mrc", self)
            return

        if vg.OUTPUT_FORMAT == vg.GridFormat.CCP4:
            vg.GridIO.write_ccp4(f"{path_prefix}.ccp4", self)
            return

        if vg.OUTPUT_FORMAT == vg.GridFormat.CMAP:
            vg.GridIO.write_cmap(f"{path_prefix}.cmap", self, self.ms.molname)
            return

        if vg.OUTPUT_FORMAT == vg.GridFormat.CMAP_PACKED:
            vg.GridIO.write_cmap(folder_out / f"{self.ms.molname}.cmap", self, f"{self.ms.molname}.{title}")
            return

        raise ValueError(f"Unknown output format: {vg.OUTPUT_FORMAT}.")


# //////////////////////////////////////////////////////////////////////////////



import re

# //////////////////////////////////////////////////////////////////////////////
class ParserIni:
    """Parser for INI-like files, extracting sections and their corresponding lines."""
    _CHAR_COMMENT = '#'

    # --------------------------------------------------------------------------
    def __init__(self, path_ini):
        self._path_ini = path_ini
        self._ini_sections: dict[str, list[str]] = {}
        with open(path_ini, 'r') as file:
            self._extract_sections(file.read())


    # --------------------------------------------------------------------------
    @classmethod
    def split_line(cls, line: str, sep: str = '=') -> tuple[str, str]:
        """
        Splits a line into a key-value pair based on the specified separator.
        """
        line = line.split(cls._CHAR_COMMENT)[0].strip() # Remove comments
        pair = tuple(map(str.strip, line.split(sep)))
        if len(pair) != 2:
            raise ValueError(f"Line '{line}' does not contain '{sep}'")
        return pair


    # --------------------------------------------------------------------------
    def has(self, key: str) -> bool:
        """
        Checks if the given key exists in the sections.
        """
        return key in self._ini_sections.keys()


    # --------------------------------------------------------------------------
    def get(self, key, default = None) -> list[str]:
        """
        Returns the value for the given key, or default if the key is not found.
        """
        return self._ini_sections.get(key, default)


    # --------------------------------------------------------------------------
    def iter_lines(self, key: str):
        """
        Iterates over the lines of the section identified by key.
        Yields each line that is not empty or a comment.
        """
        for line in self._ini_sections.get(key, []):
            line = line.strip()
            if not line or line.startswith(self._CHAR_COMMENT): continue
            yield line


    # --------------------------------------------------------------------------
    def iter_splitted_lines(self, key: str, sep: str = '='):
        """
        Iterates over the lines of the section identified by key,
        splitting each line into a key-value pair based on the specified separator.
        Yields tuples of (key, value) for each line.
        """
        for line in self.iter_lines(key):
            yield self.split_line(line, sep)


    # --------------------------------------------------------------------------
    def _extract_sections(self, data: str) -> None:
        """
        Extracts all headers and their corresponding bodies from the given string.
        Each header is a substring enclosed in square brackets, e.g. [HEADER].
        The body is the text after the header's closing bracket up to the next header or end of string.
        Text preceding the first header is captured under an empty string key.
        """

        def split_body(start: int, end: int) -> list[str]:
            return data[start:end].strip().splitlines()

        matches = list(re.finditer(r"^\s*\[(\w+)\]", data, re.MULTILINE))

        if not matches: # no headers found -> treat entire data as a single section with an empty header
            self._ini_sections[''] = split_body(0, len(data))
            return

        self._ini_sections[''] = split_body(0, matches[0].start()) # capture text before the first header

        for i, match in enumerate(matches):
            header = match.group(1)
            body_start = match.end()
            body_end = matches[i + 1].start() if (i + 1 < len(matches)) else len(data)
            if header in self._ini_sections: raise ValueError(f"Duplicate header found: {header}")
            self._ini_sections[header] = split_body(body_start, body_end)


# //////////////////////////////////////////////////////////////////////////////



import numpy as np
import MDAnalysis as mda
from pathlib import Path

import volgrids as vg

# //////////////////////////////////////////////////////////////////////////////
class MolSystem:
    def __init__(self,
        path_struct: Path = None, path_traj: Path = None,
        box_data: dict = None
    ):
        self.minCoords  : np.ndarray[float]   # minimum coordinates of the bounding box
        self.maxCoords  : np.ndarray[float]   # maximum coordinates of the bounding box
        self.resolution : np.ndarray[int]     # number of grid points in each dimension
        self.deltas     : np.ndarray[float]   # size of each grid point in each dimension
        self.cog        : np.ndarray[float]   # center of geometry of the bounding box
        self.radius     : float               # (maximum) radius of the bounding box
        self.molname    : str                 # name of the molecule
        self.do_traj    : None | bool         # whether this is a trajectory or a single structure (None if no structure is provided)
        self.system     : None | mda.Universe # MDAnalysis Universe object for the molecular system
        self.frame      : None | int          # current frame number (if trajectory is used)

        if path_struct is not None:
            ### molecular system with a molecular structure (optionally a trajectory)
            ### the bounding box values are calculated from the structure
            self._init_attrs_from_molecules(path_struct, path_traj)

        else:
            ### simple molecular system with no molecular structure
            ### the bounding box values are to be expected instead
            if box_data is None:
                raise ValueError("Either 'path_struct' or 'box_data' must be provided.")
            self._init_attrs_from_box_data(box_data)


    # --------------------------------------------------------------------------
    @classmethod
    def from_box_data(cls,
        resolution: np.ndarray, origin: np.ndarray, deltas: np.ndarray,
        molname: str = "grid"
    ) -> "MolSystem":
        """
        Create a MolSystem instance from box data. 'maxCoords' is inferred
        :param resolution: The resolution of the grid (number of points in each dimension).
        :param origin: The origin of the grid (minimum coordinates of the bounding box).
        :param deltas: The size of each grid point in each dimension.
        :param molname: The name of the molecule (default is "grid").
        :return: An instance of MolSystem with the provided box data.
        """
        return cls(box_data = {
            "resolution": resolution,
            "minCoords": origin,
            "maxCoords": origin + deltas * resolution,
            "deltas": deltas,
            "molname": molname
        })


    # --------------------------------------------------------------------------
    def _init_attrs_from_molecules(self, path_struct: Path, path_traj: Path = None):
        self.molname = path_struct.stem
        self.do_traj = path_traj is not None

        if self.do_traj:
            self.system = mda.Universe(str(path_struct), str(path_traj))
            self.frame = 0
        else:
            self.system = mda.Universe(str(path_struct))
            self.frame = None

        self._infer_box_attributes()

        self._set_deltas_resolution()

        self._warning_big_grid()


    # --------------------------------------------------------------------------
    def _init_attrs_from_box_data(self, box_data: dict):
        keys_box_data = set(box_data.keys())
        required_keys = {"molname", "minCoords", "maxCoords", "resolution", "deltas"}
        if not keys_box_data.issuperset(required_keys):
            raise ValueError(f"Box data must contain the keys: {required_keys}. Provided keys: {keys_box_data}")

        self.molname = box_data["molname"]
        self.do_traj = False

        self.system  = None
        self.frame   = None

        self.minCoords  = np.array(box_data["minCoords"],  dtype = float)
        self.maxCoords  = np.array(box_data["maxCoords"],  dtype = float)
        self.resolution = np.array(box_data["resolution"], dtype = int  )
        self.deltas     = np.array(box_data["deltas"],     dtype = float)
        self._calc_radius_and_cog()

        self._warning_big_grid()


    # --------------------------------------------------------------------------
    def _infer_box_attributes(self):
        self.minCoords = np.min(self.system.coord.positions, axis = 0) - vg.EXTRA_BOX_SIZE
        self.maxCoords = np.max(self.system.coord.positions, axis = 0) + vg.EXTRA_BOX_SIZE
        self._calc_radius_and_cog()


    # --------------------------------------------------------------------------
    def _calc_radius_and_cog(self):
        self.radius = np.linalg.norm(self.maxCoords - self.minCoords) / 2
        self.cog = (self.minCoords + self.maxCoords) / 2


    # --------------------------------------------------------------------------
    def _set_deltas_resolution(self):
        box_size = self.maxCoords - self.minCoords
        if vg.USE_FIXED_DELTAS:
            self.deltas = np.array([vg.GRID_DX, vg.GRID_DY, vg.GRID_DZ])
            self.resolution = np.round(box_size / self.deltas).astype(int)
        else:
            self.resolution = np.array([vg.GRID_XRES, vg.GRID_YRES, vg.GRID_ZRES], dtype = int)
            self.deltas = box_size / self.resolution


    # --------------------------------------------------------------------------
    def _warning_big_grid(self):
        rx, ry, rz = self.resolution
        grid_size = rx*ry*rz
        if grid_size > vg.WARNING_GRID_SIZE:
            print()
            while True:
                choice = input(
                    f">>> WARNING: resulting ({rx}x{ry}x{rz}) grid would contain {grid_size/1e6:.2f} million points. Proceed? [Y/N]\n"
                ).upper()
                if choice == 'Y': break
                if choice == 'N': exit()

    # --------------------------------------------------------------------------
    def get_relevant_atoms(self):
        raise NotImplementedError()


# //////////////////////////////////////////////////////////////////////////////



import os
from pathlib import Path
from abc import ABC, abstractmethod

# //////////////////////////////////////////////////////////////////////////////
class ParamHandler(ABC):
    class InvalidPathError(Exception): pass
    class InvalidParamError(Exception): pass
    class MissingParamError(Exception): pass
    class MissingArgsError(Exception): pass

    # --------------------------------------------------------------------------
    def __init__(self, *params_pos: str, **params_kwd: list[str]):
        self._params_pos = params_pos
        self._params_kwd = params_kwd
        self._help_str: str = ""


    # --------------------------------------------------------------------------
    @abstractmethod
    def assign_globals(self):
        return


    # --------------------------------------------------------------------------
    @property
    @abstractmethod
    def _EXPECTED_CLI_FLAGS() -> dict[str, tuple[str]]:
        raise NotImplementedError()


    # --------------------------------------------------------------------------
    @classmethod
    def parse_cli_args(cls, argv) -> tuple[list[str], dict[str, list[str]]]:
        """_EXPECTED_CLI_FLAGS is a dict where keys are flag identifiers, each associated with a list of aliases for said flag.
        This method then outputs a dict where the keys are the flag identifiers actually found in self._args,
        together with their correspondant values."""

        if cls._EXPECTED_CLI_FLAGS is None:
            raise NotImplementedError("The _EXPECTED_CLI_FLAGS attribute must be defined in the subclass.")

        alias_to_flagname = {}
        for name,aliases in cls._EXPECTED_CLI_FLAGS.items():
            alias_to_flagname = {**alias_to_flagname, **{alias:name for alias in aliases}}

        current_name = '' # '' is used for options at the start that are not associated with any flag
        params_kwd: dict[str, list[str]] = {current_name: []}

        for arg in argv:
            if arg.lower() in alias_to_flagname: # arg is a flag
                current_name = alias_to_flagname[arg.lower()]
                if current_name in params_kwd:
                    raise ValueError(f"Flag '{current_name}' is defined multiple times in the arguments.")
                params_kwd[current_name] = []

            else: # arg is a flag's option
                params_kwd[current_name].append(arg)

        params_pos = params_kwd.pop('')
        return params_pos, params_kwd


    # --------------------------------------------------------------------------
    def _set_help_str(self, *lines: str) -> None:
        self._help_str = '\n'.join(lines)


    # --------------------------------------------------------------------------
    def _exit_with_help(self, cls_error: type[Exception] = ValueError, err_msg: str = '') -> None:
        if not err_msg: # assume exit code 0 when no error message is provided
            print(f"{self._help_str}")
            exit(0)

        raise cls_error(f"\n{self._help_str}\n\nError while parsing CLI arguments:\n{err_msg}")


    # --------------------------------------------------------------------------
    def _has_params_pos(self) -> bool:
        return len(self._params_pos) > 0


    # --------------------------------------------------------------------------
    def _has_param_kwds(self, *names: str) -> bool:
        return all(name in self._params_kwd for name in names)


    # --------------------------------------------------------------------------
    def _safe_idx(self, lst: list, idx: int, err_msg: str) -> str:
        if idx < 0: idx += len(lst)

        try:
            n_elements = len(lst)
        except TypeError:
            if idx == 0:
                return lst
        else:
            if idx < n_elements:
                return lst[idx]

        self._exit_with_help(err_msg)


    # --------------------------------------------------------------------------
    def _safe_map_value(self, key: str, **values):
        value = values.get(key, None)
        if value is None:
            self._exit_with_help(self.InvalidParamError, f"Invalid parameter '{key}'. Expected one of the following: {', '.join(values.keys())}.")
        return value


    # --------------------------------------------------------------------------
    def _safe_get_param_pos(self, idx: int, err_msg: str = "") -> str:
        return self._safe_idx(self._params_pos, idx, err_msg)


    # --------------------------------------------------------------------------
    def _safe_get_param_kwd_list(self, name: str, min_len: int = 1) -> list[str]:
        if not self._has_param_kwds(name):
            self._exit_with_help(self.MissingParamError, f"The flag '{name}' was not provided.")
        lst = self._params_kwd[name]
        if len(lst) < min_len:
            self._exit_with_help(self.MissingArgsError, f"The flag '{name}' was used but not enough values were provided. At least {min_len} value(s) expected.")
        return lst


    # --------------------------------------------------------------------------
    def _safe_get_param_kwd(self, name: str, required = False, default: str = None) -> str:
        if self._has_param_kwds(name):
            lst = self._params_kwd[name]
            if lst: return lst[0]
            self._exit_with_help(self.MissingArgsError, f"The flag '{name}' was used but no value was provided.")

        if required:
            self._exit_with_help(self.MissingParamError, f"The mandatory flag '{name}' was not provided.")

        return default


    # --------------------------------------------------------------------------
    def _safe_path_file_in(self, path: str) -> Path:
        obj = Path(path)
        if not obj.exists():
            self._exit_with_help(self.InvalidPathError, f"The specified file path '{path}' does not exist.")
        if obj.is_dir():
            self._exit_with_help(self.InvalidPathError, f"The specified file path '{path}' is a folder.")
        return obj


    # --------------------------------------------------------------------------
    def _safe_path_file_out(self, path: str) -> Path:
        obj = Path(path)
        if obj.is_dir():
            self._exit_with_help(self.InvalidPathError, f"The specified file path '{path}' is a folder.")
        os.makedirs(obj.parent, exist_ok = True)
        return obj


    # --------------------------------------------------------------------------
    def _safe_path_folder_out(self, path: str) -> Path:
        obj = Path(path)
        if obj.is_file():
            self._exit_with_help(self.InvalidPathError, f"The specified folder path '{path}' is a file.")
        os.makedirs(obj, exist_ok = True)
        return obj


    # --------------------------------------------------------------------------
    def _safe_kwd_file_in(self, name: str, required = False, default: str = None) -> Path | None:
        path = self._safe_get_param_kwd(name, required, default)
        if path is None: return
        return self._safe_path_file_in(path)


    # --------------------------------------------------------------------------
    def _safe_kwd_file_out(self, name: str, required = False, default: str = None) -> Path | None:
        path = self._safe_get_param_kwd(name, required, default)
        if path is None: return
        return self._safe_path_file_out(path)


    # --------------------------------------------------------------------------
    def _safe_kwd_folder_out(self, name: str, required = False, default: str = None) -> Path | None:
        path = self._safe_get_param_kwd(name, required, default)
        if path is None: return
        return self._safe_path_folder_out(path)


    # --------------------------------------------------------------------------
    def _safe_kwd_float(self, name: str, required = False, default: float = 0.0) -> float:
        val_str: str = self._safe_get_param_kwd(name, required)
        if val_str is None: return default
        try:
            return float(val_str)
        except ValueError:
            self._exit_with_help(self.InvalidParamError, f"The value for the flag '{name}' must be a float. Got '{val_str}' instead.")


# //////////////////////////////////////////////////////////////////////////////



from ._core.mol_system import MolType, MolSystemSmiffer
from ._core.trimmer import Trimmer

from ._parsers.parser_chem_table import ParserChemTable

from ._smifs.smif import Smif
from ._smifs.apbs import SmifAPBS
from ._smifs._hbonds.hbaccepts import SmifHBAccepts
from ._smifs._hbonds.hbdonors import SmifHBDonors
from ._smifs._hydro.hydrophilic import SmifHydrophilic
from ._smifs._hydro.hydrophobic import SmifHydrophobic
from ._smifs.stacking import SmifStacking

from ._ui.param_handler import ParamHandlerSmiffer
from ._ui.app import AppSmiffer


############################# CONFIG FILE GLOBALS ##############################
import numpy as _np
_keys_other = set(globals().keys())

DO_SMIF_STACKING:    bool = True
DO_SMIF_HBA:         bool = True
DO_SMIF_HBD:         bool = True
DO_SMIF_HYDROPHOBIC: bool = True
DO_SMIF_HYDROPHILIC: bool = True
DO_SMIF_APBS:        bool = True

DO_SMIF_LOG_APBS:  bool = False
DO_SMIF_HYDRODIFF: bool = False

DO_TRIMMING_SPHERE:    bool = True
DO_TRIMMING_OCCUPANCY: bool = True
DO_TRIMMING_RNDS:      bool = True
DO_TRIMMING_FARAWAY:   bool = True
SAVE_TRIMMING_MASK:    bool = False

USE_STRUCTURE_HYDROGENS: bool = True

TRIMMING_DIST_SMALL: float = 2.5
TRIMMING_DIST_MID:   float = 3.0
TRIMMING_DIST_LARGE: float = 3.5

MAX_RNDS_DIST:   float = _np.inf
COG_CUBE_RADIUS: int = 4

TRIM_FARAWAY_DIST: float = 7.0

ENERGY_SCALE: float = 3.5

MU_HYDROPHOBIC:    float = 4.4
SIGMA_HYDROPHOBIC: float = 0.3

MU_HYDROPHILIC:    float = 3.0
SIGMA_HYDROPHILIC: float = 0.15

MU_ANGLE_HBA:    float = 129.9
MU_DIST_HBA:     float = 2.93
SIGMA_ANGLE_HBA: float = 20.0
SIGMA_DIST_HBA:  float = 0.21

MU_ANGLE_HBD_FREE:    float = 109.0
MU_DIST_HBD_FREE:     float = 2.93
SIGMA_ANGLE_HBD_FREE: float = 20.0
SIGMA_DIST_HBD_FREE:  float = 0.21

MU_ANGLE_HBD_FIXED:    float = 180.0
MU_DIST_HBD_FIXED:     float = 2.93
SIGMA_ANGLE_HBD_FIXED: float = 30.0
SIGMA_DIST_HBD_FIXED:  float = 0.21

MU_ANGLE_STACKING: float = 29.97767535
MU_DIST_STACKING:  float = 4.1876158
COV_STACKING_00:   float = 169.9862228
COV_STACKING_01:   float = 6.62318852
COV_STACKING_10:   float = 6.62318852
COV_STACKING_11:   float = 0.37123882


GAUSSIAN_KERNEL_SIGMAS: int = 4
APBS_MIN_CUTOFF: int = -2
APBS_MAX_CUTOFF: int = 3

__config_keys__ = set(globals().keys()) - _keys_other


######################## COMMAND LINE ARGUMENTS GLOBALS ########################
### These are global variables that are to be set by
### an instance of ParamHandler (or its inherited classes)

import pathlib as _pathlib
PATH_STRUCT: _pathlib.Path = None # "path/input/struct.pdb"
PATH_TRAJ:   _pathlib.Path = None # "path/input/traj.xtc"
PATH_APBS:   _pathlib.Path = None # "path/input/apbs.pqr.dx"
PATH_TABLE:  _pathlib.Path = None # "path/input/table.chem"
FOLDER_OUT:  _pathlib.Path = None # "folder/output/"

SPHERE_INFO: tuple[float, float, float, float] = None # pocket sphere info: [x, y, z, radius]
CURRENT_MOLTYPE: MolType = MolType.NONE               # type of the current molecule


############################### RUNTIME GLOBALS ################################
import volgrids as _vg
PARAMS_HBA:       _vg.ParamsGaussianBivariate
PARAMS_HBD_FREE:  _vg.ParamsGaussianBivariate
PARAMS_HBD_FIXED: _vg.ParamsGaussianBivariate
PARAMS_HPHOB:     _vg.ParamsGaussianUnivariate
PARAMS_HPHIL:     _vg.ParamsGaussianUnivariate
PARAMS_STACK:     _vg.ParamsGaussianBivariate
SIGMA_DIST_STACKING: float

APBS_ELAPSED_TIME: float = 0.0



import numpy as np

import volgrids as vg
import volgrids.smiffer as sm

# //////////////////////////////////////////////////////////////////////////////
class AppSmiffer(vg.App):
    CONFIG_MODULES = (vg, sm)
    _CLASS_PARAM_HANDLER = sm.ParamHandlerSmiffer

    _CLASS_TRIMMER = sm.Trimmer
    _CLASS_MOL_SYSTEM = sm.MolSystemSmiffer

    # --------------------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._init_globals()

        self.ms: sm.MolSystemSmiffer = self._CLASS_MOL_SYSTEM(sm.PATH_STRUCT, sm.PATH_TRAJ)
        self.trimmer: sm.Trimmer = self._CLASS_TRIMMER.init_infer_dists(self.ms)
        self.timer = vg.Timer(
            f">>> Now processing {sm.CURRENT_MOLTYPE.name:>4} '{self.ms.molname}'"+\
            f" in '{'PocketSphere' if self.ms.do_ps else 'Whole'}' mode"
        )


    # --------------------------------------------------------------------------
    def run(self):
        if sm.CURRENT_MOLTYPE.is_ligand():
            sm.DO_SMIF_APBS = False
            print("\n...--- ligand: skipping APBS SMIF calculation.", end = ' ', flush = True)

        self.timer.start()

        if self.ms.do_traj: # TRAJECTORY MODE
            print()
            for _ in self.ms.system.trajectory:
                self.ms.frame += 1
                timer_frame = vg.Timer(f"...>>> Frame {self.ms.frame}/{len(self.ms.system.trajectory)}")
                timer_frame.start()
                self._process_grids()
                timer_frame.end()

        else: # SINGLE PDB MODE
            self._process_grids()

        self.timer.end(text = "SMIFS", minus = sm.APBS_ELAPSED_TIME)


    # --------------------------------------------------------------------------
    def _import_config_dependencies(self):
        return {"np": np, "vg": vg}


    # --------------------------------------------------------------------------
    def _init_globals(self):
        sm.PARAMS_HPHOB = vg.ParamsGaussianUnivariate(
            mu = sm.MU_HYDROPHOBIC, sigma = sm.SIGMA_HYDROPHOBIC,
        )
        sm.PARAMS_HPHIL = vg.ParamsGaussianUnivariate(
            mu = sm.MU_HYDROPHILIC, sigma = sm.SIGMA_HYDROPHILIC,
        )
        sm.PARAMS_HBA = vg.ParamsGaussianBivariate(
            mu_0 = sm.MU_ANGLE_HBA, mu_1 = sm.MU_DIST_HBA,
            cov_00 = sm.SIGMA_ANGLE_HBA**2, cov_01 = 0,
            cov_10 = 0,  cov_11 = sm.SIGMA_DIST_HBA**2,
        )
        sm.PARAMS_HBD_FREE = vg.ParamsGaussianBivariate(
            mu_0 = sm.MU_ANGLE_HBD_FREE, mu_1 = sm.MU_DIST_HBD_FREE,
            cov_00 = sm.SIGMA_ANGLE_HBD_FREE**2, cov_01 = 0,
            cov_10 = 0,  cov_11 = sm.SIGMA_DIST_HBD_FREE**2,
        )
        sm.PARAMS_HBD_FIXED = vg.ParamsGaussianBivariate(
            mu_0 = sm.MU_ANGLE_HBD_FIXED, mu_1 = sm.MU_DIST_HBD_FIXED,
            cov_00 = sm.SIGMA_ANGLE_HBD_FIXED**2, cov_01 = 0,
            cov_10 = 0,  cov_11 = sm.SIGMA_DIST_HBD_FIXED**2,
        )
        sm.PARAMS_STACK = vg.ParamsGaussianBivariate(
            mu_0 = sm.MU_ANGLE_STACKING, mu_1 = sm.MU_DIST_STACKING,
            cov_00 = sm.COV_STACKING_00, cov_01 = sm.COV_STACKING_01,
            cov_10 = sm.COV_STACKING_10, cov_11 = sm.COV_STACKING_11,
        )

        ### square root of the DIST contribution to sm.COV_STACKING,
        sm.SIGMA_DIST_STACKING = np.sqrt(sm.COV_STACKING_11)


    # --------------------------------------------------------------------------
    def _process_grids(self):
        ### APBS must be calculated first and split into two parts,
        ### because it can potentially set vg.PQR_CONTENTS_TEMP (used for trimming)
        if sm.DO_SMIF_APBS:
            smif_apbs: sm.SmifAPBS = self._calc_smif(sm.SmifAPBS)
            if vg.PQR_CONTENTS_TEMP:
                self.trimmer.ms = sm.MolSystemSmiffer.from_pqr_data(vg.PQR_CONTENTS_TEMP)

        self.trimmer.trim()

        if sm.DO_SMIF_APBS:
            smif_apbs.reshape_as_other(self.trimmer.specific_masks["large"])
            self._trim_and_save_smif(
                smif_apbs, key_trimming = "large", title = "apbs"
            )
            del self.trimmer.specific_masks["large"]


        ### Calculate standard SMIF grids
        if sm.DO_SMIF_HYDROPHILIC:
            smif_hphil = self._calc_smif(sm.SmifHydrophilic)
            self._trim_and_save_smif(
                smif_hphil, key_trimming = "small", title = "hydrophilic"
            )
            del self.trimmer.specific_masks["small"]

        if sm.DO_SMIF_HYDROPHOBIC:
            smif_hphob = self._calc_smif(sm.SmifHydrophobic)
            self._trim_and_save_smif(
                smif_hphob, key_trimming = "mid", title = "hydrophobic"
            )

        if sm.DO_SMIF_HBA:
            self._trim_and_save_smif(
                self._calc_smif(sm.SmifHBAccepts),
                key_trimming = "mid", title = "hbacceptors"
            )

        if sm.DO_SMIF_HBD:
            self._trim_and_save_smif(
                self._calc_smif(sm.SmifHBDonors),
                key_trimming = "mid", title = "hbdonors"
            )

        if sm.DO_SMIF_STACKING:
            self._trim_and_save_smif(
                self._calc_smif(sm.SmifStacking),
                key_trimming = "mid", title = "stacking"
            )


        ### Calculate additional grids
        if sm.SAVE_TRIMMING_MASK:
            mask = self.trimmer.get_mask("mid")
            reverse = vg.Grid.reverse(mask) # save the points that are NOT trimmed
            reverse.save_data(sm.FOLDER_OUT, f"trimming")

        if sm.DO_SMIF_HYDROPHOBIC and sm.DO_SMIF_HYDROPHILIC and sm.DO_SMIF_HYDRODIFF:
            grid_hpdiff = smif_hphob - smif_hphil
            grid_hpdiff.save_data(sm.FOLDER_OUT, "hydrodiff")

        if sm.DO_SMIF_LOG_APBS:
            smif_apbs.apply_logabs_transform()
            smif_apbs.save_data(sm.FOLDER_OUT, "apbslog")

        if not self.ms.do_traj and vg.PQR_CONTENTS_TEMP:
            path_pqr = sm.FOLDER_OUT / f"{self.ms.molname}.pqr"
            path_pqr.write_text(vg.PQR_CONTENTS_TEMP)


    # --------------------------------------------------------------------------
    def _calc_smif(self, cls_smif: type[sm.Smif]) -> "sm.Smif":
        smif: sm.Smif = cls_smif(self.ms)
        smif.populate_grid()
        return smif


    # --------------------------------------------------------------------------
    def _trim_and_save_smif(self, smif: sm.Smif, key_trimming: str, title: str) -> None:
        self.trimmer.mask_grid(smif, key_trimming)
        smif.save_data(sm.FOLDER_OUT, title)


# //////////////////////////////////////////////////////////////////////////////



import tempfile
import numpy as np
from pathlib import Path
from enum import Enum, auto

import volgrids as vg
import volgrids.smiffer as sm

# //////////////////////////////////////////////////////////////////////////////
class MolType(Enum):
    NONE = auto()
    PROT = auto()
    RNA = auto()
    LIGAND = auto()

    # --------------------------------------------------------------------------
    def is_none(self):   return self == MolType.NONE
    def is_prot(self):   return self == MolType.PROT
    def is_rna(self):    return self == MolType.RNA
    def is_ligand(self): return self == MolType.LIGAND


# //////////////////////////////////////////////////////////////////////////////
class MolSystemSmiffer(vg.MolSystem):
    def __init__(self, path_struct: Path, path_traj: Path = None):
        self.do_ps = sm.SPHERE_INFO is not None
        self.chemtable = sm.ParserChemTable(self._get_path_table())
        self._init_attrs_from_molecules(path_struct, path_traj)


    # --------------------------------------------------------------------------
    @classmethod
    def from_pqr_data(cls, pqr_data: str):
        with tempfile.NamedTemporaryFile(mode = "w+", suffix = ".pqr", delete = True) as tmp_pqr:
            tmp_pqr.write(pqr_data)
            tmp_pqr.flush()
            return cls(Path(tmp_pqr.name), None)


    # --------------------------------------------------------------------------
    def get_relevant_atoms(self):
        if self.do_ps:
            xcog, ycog, zcog, radius = sm.SPHERE_INFO
            return self.system.select_atoms(
                f"{self.chemtable.selection_query} and point {xcog} {ycog} {zcog} {radius}"
            )

        return self.system.select_atoms(self.chemtable.selection_query)


    # --------------------------------------------------------------------------
    def get_relevant_atoms_broad(self, trimming_dist):
        if self.do_ps:
            xcog, ycog, zcog, radius = sm.SPHERE_INFO
            return self.system.select_atoms(
                f"{self.chemtable.selection_query} and point {xcog} {ycog} {zcog} {radius + trimming_dist}"
            )

        return self.system.select_atoms(self.chemtable.selection_query)


    # --------------------------------------------------------------------------
    def _infer_box_attributes(self):
        if self.do_ps:
            xcog, ycog, zcog, radius = sm.SPHERE_INFO
            self.cog = np.array([xcog, ycog, zcog])
            self.minCoords = self.cog - radius
            self.maxCoords = self.cog + radius
            self.radius = radius
            return

        super()._infer_box_attributes()


    # --------------------------------------------------------------------------
    def _get_path_table(self) -> Path:
        if sm.PATH_TABLE: return sm.PATH_TABLE

        folder_default_tables = Path("_tables")

        if sm.CURRENT_MOLTYPE == MolType.PROT:
            return vg.resolve_path_package(folder_default_tables / "prot.chem")

        if sm.CURRENT_MOLTYPE == MolType.RNA:
            return vg.resolve_path_package(folder_default_tables / "rna.chem")

        raise ValueError(f"No default table for the specified molecular type '{sm.CURRENT_MOLTYPE}'. Please provide a path to a custom table.")


# //////////////////////////////////////////////////////////////////////////////



import volgrids as vg
import volgrids.smiffer as sm

# //////////////////////////////////////////////////////////////////////////////
class ParamHandlerSmiffer(vg.ParamHandler):
    _EXPECTED_CLI_FLAGS = {
        "help"  : ("-h", "--help"),
        "output": ("-o", "--output"),
        "traj"  : ("-t", "--traj"),
        "apbs"  : ("-a", "--apbs"),
        "sphere": ("-s", "--sphere"),
        "table" : ("-b", "--table"),
        "config": ("-c", "--config"),
    }


    # --------------------------------------------------------------------------
    def assign_globals(self):
        self._set_help_str(
            "usage: volgrids smiffer [prot|rna|convert|pack|unpack] [options...]",
            "Available modes:",
            "  prot     - Calculate SMIFs for protein structures.",
            "  rna      - Calculate SMIFs for RNA structures.",
            "  ligand   - Calculate SMIFs for ligand structures. A .chem table must be provided.",
            "Run 'volgrids smiffer [mode] --help' for more details on each mode.",
        )
        if self._has_param_kwds("help") and not self._has_params_pos():
            self._exit_with_help()

        mode = self._safe_get_param_pos(0)
        sm.CURRENT_MOLTYPE = self._safe_map_value(mode.lower(),
            prot = sm.MolType.PROT,
            rna = sm.MolType.RNA,
            ligand = sm.MolType.LIGAND,
        )

        self._set_help_str(
            f"usage: volgrids smiffer {mode} [path/input/struct.pdb] [options...]",
            "Available options:",
            "-h, --help        Show this help message and exit.",
            "-o, --output      Folder path where the output SMIFs should be stored. If not provided, the parent folder of the input structure file will be used.",
            "-t, --traj        File path to a trajectory file (e.g. XTC) supported by MDAnalysis. Activates 'traj' mode: calculate SMIFs for all the frames and save them as a CMAP-series file.",
            "-a, --apbs        File path to a cached output of APBS for the respective structure file. Prevents the automatic calculation of APBS performed by volgrids' smiffer.",
            "-b, --table       File path to a .chem table file to use for ligand mode, or to override the default macromolecules' tables.",
            "-c, --config      File path to a configuration file with global settings, to override the default settings (e.g. config_volgrids.ini).",
            "-s, --sphere      Activate 'pocket sphere' mode by providing the X, Y, Z coordinates (sphere center) and the sphere radius R for a sphere. If not provided, 'whole' mode is assumed.",
        )
        if self._has_param_kwds("help"):
            self._exit_with_help()

        sm.PATH_STRUCT = self._safe_path_file_in(
            self._safe_get_param_pos(1,
               err_msg = "No input structure file provided. Provide a path to the structure file as first positional argument."
            )
        )

        sm.FOLDER_OUT = self._safe_kwd_folder_out("output", default = sm.PATH_STRUCT.parent)
        sm.PATH_APBS  = self._safe_kwd_file_in("apbs")
        sm.PATH_TRAJ  = self._safe_kwd_file_in("traj")
        sm.PATH_TABLE = self._safe_kwd_file_in("table")

        self._handle_params_configs()
        self._handle_params_sphere()
        self._assert_traj_apbs()
        self._assert_ligand_has_table()


    # --------------------------------------------------------------------------
    def _handle_params_sphere(self):
        if not self._has_param_kwds("sphere"): return
        params_sphere = self._params_kwd["sphere"]
        try:
            x_cog  = float(self._safe_idx(params_sphere, 0, "Missing sphere center X coordinate."))
            y_cog  = float(self._safe_idx(params_sphere, 1, "Missing sphere center Y coordinate."))
            z_cog  = float(self._safe_idx(params_sphere, 2, "Missing sphere center Z coordinate."))
            radius = float(self._safe_idx(params_sphere, 3, "Missing sphere radius."))
        except ValueError:
            self._exit_with_help(self.InvalidParamError, "Sphere options must be numeric values.")
        sm.SPHERE_INFO = (x_cog, y_cog, z_cog, radius)


    # --------------------------------------------------------------------------
    def _handle_params_configs(self):
        path_config = self._safe_kwd_file_in("config") # [WIP]
        if path_config is None: return
        vg.PATHS_CUSTOM_CONFIG = [path_config]


    # --------------------------------------------------------------------------
    def _assert_traj_apbs(self):
        if not sm.DO_SMIF_APBS: return
        if (sm.PATH_TRAJ is None) or (sm.PATH_APBS is None): return
        raise self.InvalidParamError(
            f"The APBS output '{sm.PATH_APBS}' was provided. However, "+
            "trajectory mode is enabled, so this file would be ambiguous. "+
            "Please either disable trajectory mode or remove the APBS file input. "
        )


    # --------------------------------------------------------------------------
    def _assert_ligand_has_table(self):
        if not sm.CURRENT_MOLTYPE.is_ligand(): return
        if self._has_param_kwds("table"): return
        self._exit_with_help(
            self.MissingParamError,
            "No table file provided for ligand mode. Use -b or --table to specify the path to the .chem table file."
        )


# //////////////////////////////////////////////////////////////////////////////



import warnings
from abc import ABC
import MDAnalysis as mda

import volgrids as vg
import volgrids.smiffer as sm

from .hb import SmifHBonds
from .triplet import Triplet

# ------------------------------------------------------------------------------
def _has_prev_res(atoms, triplet: Triplet) -> bool:
    return len(atoms.select_atoms(triplet.str_prev_res)) > 0

# ------------------------------------------------------------------------------
def _has_next_res(atoms, triplet: Triplet) -> bool:
    return len(atoms.select_atoms(triplet.str_next_res)) > 0


# //////////////////////////////////////////////////////////////////////////////
class SmifHBDonors(SmifHBonds, ABC):
    # --------------------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.hbond_getter = sm.ParserChemTable.get_names_hbd
        self._kernel_hbd_free = vg.KernelGaussianBivariateAngleDist(
            radius = sm.MU_DIST_HBD_FREE + sm.GAUSSIAN_KERNEL_SIGMAS * sm.SIGMA_DIST_HBD_FREE,
            deltas = self.ms.deltas, dtype = vg.FLOAT_DTYPE, params = sm.PARAMS_HBD_FREE
        )
        self._kernel_hbd_free.link_to_grid(self.grid, self.ms.minCoords)

        self._kernel_hbd_fixed = vg.KernelGaussianBivariateAngleDist(
            radius = sm.MU_DIST_HBD_FIXED + sm.GAUSSIAN_KERNEL_SIGMAS * sm.SIGMA_DIST_HBD_FIXED,
            deltas = self.ms.deltas, dtype = vg.FLOAT_DTYPE, params = sm.PARAMS_HBD_FIXED
        )
        self._kernel_hbd_fixed.link_to_grid(self.grid, self.ms.minCoords)


    # --------------------------------------------------------------------------
    def find_tail_head_positions(self, triplet: Triplet) -> None:
        if triplet.pos_head is not None: # head position is already set for succesful sm.USE_STRUCTURE_HYDROGENS iterations
            return

        triplet.set_pos_head(self.res_atoms)

        ############################### TAIL POSITION
        ### special cases for protein
        if sm.CURRENT_MOLTYPE == sm.MolType.PROT:
            if triplet.resname == "PRO": # donor only if there is no previous residue
                if _has_prev_res(self.all_atoms, triplet): return

            elif triplet.interactor == "N": # tail points are in different residues
                if _has_prev_res(self.all_atoms, triplet):
                    triplet.set_pos_tail_custom( # N of peptide bond
                        atoms = self.all_atoms,
                        query_t0 = triplet.str_prev_res,
                        query_t1 = triplet.str_this_res
                    )
                    self.kernel = self._kernel_hbd_fixed
                    return

                triplet.set_pos_tail_custom( # N of N-terminus
                    atoms = self.all_atoms,
                    query_t0 = f"{triplet.str_this_res} and name CA",
                    query_t1 = f"{triplet.str_this_res} and name CA"
                )
                self.kernel = self._get_relevant_kernel(triplet)
                return


        ### special cases for RNA
        if sm.CURRENT_MOLTYPE == sm.MolType.RNA:
            if triplet.interactor == "O3'": # donor only if there is no next residue
                if _has_next_res(self.all_atoms, triplet): return

            elif triplet.interactor == "O5'": # donor only if there is no previous residue
                if _has_prev_res(self.all_atoms, triplet): return

        triplet.set_pos_tail(self.res_atoms)


    # --------------------------------------------------------------------------
    def _iter_triplets(self):
        if sm.USE_STRUCTURE_HYDROGENS:
            self._attempt_to_guess_bonds()

        for triplet in super()._iter_triplets():
            if triplet.interactor in self.processed_interactors: continue

            if sm.USE_STRUCTURE_HYDROGENS:
                for hydrogen in triplet.get_interactor_bonded_hydrogens(self.res_atoms):
                    triplet.pos_tail = triplet.pos_interactor
                    triplet.pos_head = hydrogen.position
                    self.kernel = self._kernel_hbd_fixed
                    self.processed_interactors.add(triplet.interactor)
                    yield triplet

            if triplet.pos_head is None: # sm.USE_STRUCTURE_HYDROGENS falls back to "no-hydrogen" model if no hydrogens found
                self.kernel = self._get_relevant_kernel(triplet)
                yield triplet


    # --------------------------------------------------------------------------
    def _get_relevant_kernel(self, triplet: Triplet) -> vg.KernelGaussianBivariateAngleDist:
        return self._kernel_hbd_fixed if triplet.hbond_fixed else self._kernel_hbd_free


    # --------------------------------------------------------------------------
    def _attempt_to_guess_bonds(self):
        hydrogens = self.ms.system.select_atoms("name H*")
        if len(hydrogens) == 0:
            sm.USE_STRUCTURE_HYDROGENS = False
            return

        try:
            u = mda.Merge(self.all_atoms, hydrogens) # temporary universe that excludes any unwanted atoms (like ions with undefined vdw radii)...
            u.guess_TopologyAttrs(to_guess = ["bonds"]) # ... so that there are no problems with the bond guessing
        except (ValueError, AttributeError):
            warnings.warn("MDAnalysis could not guess bonds for hydrogens. Falling back to non-hydrogen model for H-bond donors.")
            sm.USE_STRUCTURE_HYDROGENS = False
            return

        ### the bonds are contained in these newly defined atomgroup, so update the all_atoms reference
        self.all_atoms = u.atoms


# //////////////////////////////////////////////////////////////////////////////



import numpy as np
from abc import ABC

import volgrids as vg

# //////////////////////////////////////////////////////////////////////////////
class KernelGaussian(vg.Kernel, ABC):
    def __init__(self, radius, deltas, dtype, params: "vg.ParamsGaussian"):
        super().__init__(radius, deltas, dtype)
        self.params = params


# //////////////////////////////////////////////////////////////////////////////
class KernelGaussianUnivariateDist(KernelGaussian):
    """For generating univariate gaussian spheres (e.g. for hydrophob)"""
    def __init__(self, radius, deltas, dtype, params: "vg.ParamsGaussianUnivariate"):
        super().__init__(radius, deltas, dtype, params)
        self.kernel = vg.Math.univariate_gaussian(self.dist, params.mu, params.sigma)


# //////////////////////////////////////////////////////////////////////////////
class KernelGaussianBivariateAngleDist(KernelGaussian):
    """For generating multivariate gaussian distributions (for hba, hbd, stacking)"""
    def recalculate_kernel(self, normal, isStacking: bool):
        self.params: "vg.ParamsGaussianBivariate"
        angles = vg.Math.get_angle(
            self.shifted_coords, normal,
            flag_corrections = "stacking" if isStacking else "hbonds"
        )
        input_mat = np.concatenate(
            (
                np.resize(angles,    list(angles.shape)    + [1]),
                np.resize(self.dist, list(self.dist.shape) + [1]),
            ),
            axis = 3
        )
        self.kernel = vg.Math.bivariate_gaussian(input_mat, self.params.mu, self.params.cov_inv)


# //////////////////////////////////////////////////////////////////////////////



import numpy as np
from scipy import interpolate

import volgrids as vg

# //////////////////////////////////////////////////////////////////////////////
class Math:
    @staticmethod
    def normalize(vector):
        """
        input:  (3,)
        output: (3,)
        """
        return vector / np.linalg.norm(vector)

    # --------------------------------------------------------------------------
    @staticmethod
    def dot_product(m_vectors, vector):
        """
        input (m_vectors): (xres, yres, zres, 3)
        input (vector):    (3,)
        output:            (xres, yres, zres)
        """
        d = m_vectors[:,:,:,0] * vector[0] + \
            m_vectors[:,:,:,1] * vector[1] + \
            m_vectors[:,:,:,2] * vector[2]
        return d

    # --------------------------------------------------------------------------
    @staticmethod
    def get_norm(m_vectors):
        """
        input:  (xres, yres, zres, 3)
        output: (xres, yres, zres)
        """
        norm = np.sqrt(
            m_vectors[:,:,:,0]**2 + \
            m_vectors[:,:,:,1]**2 + \
            m_vectors[:,:,:,2]**2
        )
        return norm

    # --------------------------------------------------------------------------
    @staticmethod
    def get_angle(m_vectors, vector, in_degrees = True, flag_corrections = ''):
        """
        input (m): (xres, yres, zres, 3)
        input (v): (3,)
        output   : (xres, yres, zres)
        """
        RIGHT_ANGLE = np.pi / 2
        SEMICIRCLE  = np.pi
        TO_DEGREES  = 180 / np.pi

        numerator = Math.dot_product(m_vectors, vector)
        denominator = Math.get_norm(m_vectors) * np.linalg.norm(vector)

        mask = denominator == 0
        numerator[mask] = 1
        denominator[mask] = 1

        cos_val = np.clip(numerator / denominator, -1, 1)
        angle = np.arccos(cos_val) # in radians

        if flag_corrections == "stacking":
            angle[angle >= RIGHT_ANGLE] = SEMICIRCLE - angle[angle >= RIGHT_ANGLE]
        elif flag_corrections == "hbonds":
            angle = SEMICIRCLE - angle

        angle[mask] = -RIGHT_ANGLE

        return angle*TO_DEGREES if in_degrees else angle

    # --------------------------------------------------------------------------
    @staticmethod
    def get_projection(m_vectors, vector):
        """
        projection of the m_vectors on the vector
        input (m): (xres, yres, zres, 3)
        input (v): (3,)
        output   : (xres, yres, zres)
        """
        dot_uv = Math.dot_product(m_vectors, vector)
        norm_v = np.linalg.norm(vector)
        return 0 if (norm_v == 0) else dot_uv / norm_v

    # --------------------------------------------------------------------------
    @staticmethod
    def get_projection_height(m_vectors, vector):
        """
        given the projection of the m_vectors on the vector, consider the "height"
        as the distance from the projection to the vector
        input (m): (xres, yres, zres, 3)
        input (v): (3,)
        output   : (xres, yres, zres)
        """

        projection = Math.get_projection(m_vectors, vector)
        norm_mvectors = Math.get_norm(m_vectors)
        height = np.sqrt(norm_mvectors**2 - projection**2)
        return height

    # --------------------------------------------------------------------------
    @staticmethod
    def univariate_gaussian(x, mu, sigma):
        """ input_mat "x" shape: (xsize, ysize, zsize), values: dist """
        u = x - mu
        s = 1 / (sigma ** 2)
        return np.exp(-(1/2) * np.power(u, 2) * s)

    # --------------------------------------------------------------------------
    @staticmethod
    def bivariate_gaussian(x, mu, cov_inv):
        """ input_mat "x" shape: (xsize, ysize, zsize, 2 = (dist, beta)) """
        u = x - mu
        sigma = cov_inv

        ux, uy = u[:,:,:,0], u[:,:,:,1]
        a,b,c,d = sigma[0,0], sigma[0,1], sigma[1,0], sigma[1,1]
        return np.exp(-(1/2) * (ux * (a * ux + b * uy) + uy * (c * ux + d * uy)))

    # --------------------------------------------------------------------------
    @staticmethod
    def interpolate_3d(x0, y0, z0, data_0, new_coords):
        return interpolate.RegularGridInterpolator(
            (x0, y0, z0), data_0, bounds_error = False, fill_value = 0
        )(new_coords).T

    # --------------------------------------------------------------------------
    @staticmethod
    def get_coords_array(resolution, deltas, minCoords = None):
        """
        input:  resolution (3,)
                deltas (3,)
                minCoords (3,)
        output: coords (xres, yres, zres, 3)
        """
        xres, yres, zres = resolution
        dx, dy, dz = deltas
        x0, y0, z0 = (0,0,0) if minCoords is None else minCoords

        xrange = x0 + np.linspace(0, dx * (xres - 1), xres)
        yrange = y0 + np.linspace(0, dy * (yres - 1), yres)
        zrange = z0 + np.linspace(0, dz * (zres - 1), zres)
        x,y,z = np.meshgrid(xrange, yrange, zrange, indexing = "ij")

        grid = np.empty((xres, yres, zres, 3), dtype = vg.FLOAT_DTYPE)
        grid[:,:,:,0] = x
        grid[:,:,:,1] = y
        grid[:,:,:,2] = z
        return grid


# //////////////////////////////////////////////////////////////////////////////



import volgrids as vg
import volgrids.smiffer as sm

from .hb import SmifHBonds
from .triplet import Triplet

# //////////////////////////////////////////////////////////////////////////////
class SmifHBAccepts(SmifHBonds):
    # --------------------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.kernel = vg.KernelGaussianBivariateAngleDist(
            radius = sm.MU_DIST_HBA + sm.GAUSSIAN_KERNEL_SIGMAS * sm.SIGMA_DIST_HBA,
            deltas = self.ms.deltas, dtype = vg.FLOAT_DTYPE, params = sm.PARAMS_HBA
        )
        self.kernel.link_to_grid(self.grid, self.ms.minCoords)
        self.hbond_getter = sm.ParserChemTable.get_names_hba


    # --------------------------------------------------------------------------
    def find_tail_head_positions(self, triplet: Triplet) -> None:
        triplet.set_pos_head(self.res_atoms)

        ############################### TAIL POSITION
        ### special cases for RNA
        if sm.CURRENT_MOLTYPE == sm.MolType.RNA:
            if triplet.interactor == "O3'": # tail points are in different residues
                triplet.set_pos_tail_custom(
                    atoms = self.all_atoms,
                    query_t0 = triplet.str_this_res,
                    query_t1 = triplet.str_next_res
                )
                return

        triplet.set_pos_tail(self.res_atoms)


# //////////////////////////////////////////////////////////////////////////////



from abc import ABC, abstractmethod
from pathlib import Path

import volgrids as vg

# //////////////////////////////////////////////////////////////////////////////
class App(ABC):
    # --------------------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        handler = self._CLASS_PARAM_HANDLER(*args, **kwargs)
        handler.assign_globals()
        self.load_configs()


    # --------------------------------------------------------------------------
    @classmethod
    def from_cli(cls, argv):
        params_pos, params_kwd = cls._CLASS_PARAM_HANDLER.parse_cli_args(argv)
        return cls(*params_pos, **params_kwd)


    # --------------------------------------------------------------------------
    @classmethod
    def from_str_argv(cls, str_argv: str):
        argv = str_argv.split()
        params_pos, params_kwd = cls._CLASS_PARAM_HANDLER.parse_cli_args(argv)
        return cls(*params_pos, **params_kwd)


    # --------------------------------------------------------------------------
    def load_configs(self):
        if vg.PATH_DEFAULT_CONFIG is not None:
            self._load_config_file(vg.PATH_DEFAULT_CONFIG)

        for path_config in vg.PATHS_CUSTOM_CONFIG:
            self._load_config_file(path_config)


    # --------------------------------------------------------------------------
    def _load_config_file(self, path_config: Path) -> None:
        parser = vg.ParserConfig(path_config)
        scope_dependencies = self._import_config_dependencies()

        all_known_keys = set(k for scope_module in self.CONFIG_MODULES for k in scope_module.__config_keys__)
        for scope_module in self.CONFIG_MODULES:
            parser.apply_config(
                scope_module = scope_module.__dict__,
                scope_dependencies = scope_dependencies,
                this_module_keys = scope_module.__config_keys__,
                all_known_keys = all_known_keys
            )


    # --------------------------------------------------------------------------
    @property
    @abstractmethod
    def CONFIG_MODULES() -> tuple:
        """class variable that links the pertinent keys from the config file, to the module to be configured.
        For example: `{"VOLGRIDS": vg, "SMIFFER": sm}`, assuming `import volgrids.vgrids as vg` and `import volgrids.smiffer as sm`."""
        raise NotImplementedError()


    # --------------------------------------------------------------------------
    @property
    @abstractmethod
    def _CLASS_PARAM_HANDLER() -> type["vg.ParamHandler"]:
        raise NotImplementedError()


    # --------------------------------------------------------------------------
    @abstractmethod
    def run(self):
        raise NotImplementedError()


    # --------------------------------------------------------------------------
    @abstractmethod
    def _import_config_dependencies(self) -> dict[str, any]:
        """Import any modules that can be used by the config file.
        For example, if a key-value pair is VALUE=np.inf, you should import numpy as np inside this method.
        Return the scope with the imported modules, for example by adding "return locals()" at the end.
        """
        raise NotImplementedError()


# //////////////////////////////////////////////////////////////////////////////



import os, h5py
import numpy as np
import gridData as gd
from pathlib import Path
from enum import Enum, auto

import volgrids as vg

# //////////////////////////////////////////////////////////////////////////////
class GridFormat(Enum):
    DX = auto()
    MRC = auto()
    CCP4 = auto()
    CMAP = auto()
    CMAP_PACKED = auto()


# //////////////////////////////////////////////////////////////////////////////
class GridIO:
    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ MAIN I/O OPERATIONS
    @staticmethod
    def read_dx(path_dx) -> "vg.Grid":
        parser_dx = gd.Grid(path_dx)
        ms = vg.MolSystem.from_box_data(
            resolution = parser_dx.grid.shape,
            origin = parser_dx.origin,
            deltas = parser_dx.delta
        )
        obj = vg.Grid(ms, init_grid = False)
        obj.grid = parser_dx.grid
        obj.fmt = vg.GridFormat.DX
        return obj


    # --------------------------------------------------------------------------
    @staticmethod
    def read_mrc(path_mrc) -> "vg.Grid":
        with gd.mrc.mrcfile.open(path_mrc) as parser:
            ##### assume that MRC always follows the origin follows the "real space" MRC convention
            orig = parser.header["origin"]
            used_origin = np.array([orig['x'], orig['y'], orig['z']])

        obj = _read_mrc_ccp4(path_mrc, used_origin)
        obj.fmt = vg.GridFormat.MRC
        return obj


    # --------------------------------------------------------------------------
    @staticmethod
    def read_ccp4(path_ccp4) -> "vg.Grid":
        with gd.mrc.mrcfile.open(path_ccp4) as parser:
            orig = parser.header["origin"]
            if (orig['x'] == 0.0 and orig['y'] == 0.0 and orig['z'] == 0.0):
                ##### assume the origin follows the "integer offset" CCP4 convention, so use that one
                ##### https://mail.cgl.ucsf.edu/mailman/archives/list/chimera-users@cgl.ucsf.edu/thread/GLP62Q2WIBJ6ZU4H6BWXWVXETYSXVIWS/
                used_origin = np.array([
                    parser.voxel_size['x'] * parser.header["nxstart"],
                    parser.voxel_size['y'] * parser.header["nystart"],
                    parser.voxel_size['z'] * parser.header["nzstart"],
                ])
            else:
                ##### assume the origin follows the "real space" MRC convention, so use that one
                used_origin = np.array([orig['x'], orig['y'], orig['z']])

        obj = _read_mrc_ccp4(path_ccp4, used_origin)
        obj.fmt = vg.GridFormat.CCP4
        return obj


    # --------------------------------------------------------------------------
    @staticmethod
    def read_cmap(path_cmap, key) -> "vg.Grid":
        with h5py.File(path_cmap, 'r') as parser:
            frame = parser["Chimera"][key]
            rz, ry, rx = frame["data_zyx"].shape
            ox, oy, oz = frame.attrs["origin"]
            dz, dy, dx = frame.attrs["step"]
            ms = vg.MolSystem.from_box_data(
                resolution = np.array([rx, ry, rz]),
                origin = np.array([ox, oy, oz]),
                deltas = np.array([dx, dy, dz])
            )
            obj = vg.Grid(ms, init_grid = False)
            obj.grid = frame["data_zyx"][()].transpose(2,1,0)

            n_keys = len(parser["Chimera"].keys())
            obj.fmt = vg.GridFormat.CMAP_PACKED \
                if (n_keys > 1) else vg.GridFormat.CMAP
        return obj


    # --------------------------------------------------------------------------
    @staticmethod
    def write_dx(path_dx, data: "vg.Grid"):
        ints = (int, np.int8, np.int16, np.int32, np.int64)
        floats = (float, np.float16, np.float32, np.float64)

        if data.grid.dtype in floats:
            grid_data = data.grid
            dtype = '"float"'
            fmt = "%.3f"
        elif data.grid.dtype in ints:
            grid_data = data.grid
            dtype = '"int"'
            fmt = "%i"
        elif data.grid.dtype == bool:
            grid_data = data.grid.astype(int)
            dtype = '"int"'
            fmt = "%i"
        else:
            raise TypeError(f"Unsupported data type for DX output: {data.grid.dtype}")

        header = '\n'.join((
            "# OpenDX density file written by volgrids",
            "# File format: http://opendx.sdsc.edu/docs/html/pages/usrgu068.htm#HDREDF",
            "# Data are embedded in the header and tied to the grid positions.",
            "# Data is written in C array order: In grid[x,y,z] the axis z is fastest",
            "# varying, then y, then finally x, i.e. z is the innermost loop.",
            f"object 1 class gridpositions counts {data.xres} {data.yres} {data.zres}",
            f"origin {data.xmin:6e} {data.ymin:6e} {data.zmin:6e}",
            f"delta {data.dx:6e} {0:6e} {0:6e}",
            f"delta {0:6e} {data.dy:6e} {0:6e}",
            f"delta {0:6e} {0:6e} {data.dz:6e}",
            f"object 2 class gridconnections counts  {data.xres} {data.yres} {data.zres}",
            f"object 3 class array type {dtype} rank 0 items {data.xres*data.yres*data.zres}, data follows",
        ))
        footer = '\n'.join((
            '',
            'attribute "dep" string "positions"',
            'object "density" class field',
            'component "positions" value 1',
            'component "connections" value 2',
            'component "data" value 3',
        ))

        ########### reshape the grid array
        grid_size = np.prod(grid_data.shape)
        dx_rows = grid_size // 3

        truncated_arr, extra_arr = np.split(grid_data.flatten(), [3*dx_rows])
        data_out = truncated_arr.reshape(dx_rows, 3)
        last_row = extra_arr.reshape(1, len(extra_arr))

        ########### export reshaped data
        with open(path_dx, "wb") as file:
            np.savetxt(
                file, data_out, fmt = fmt, delimiter = '\t',
                header = header, comments = ''
            )
            np.savetxt(
                file, last_row, fmt = fmt, delimiter = '\t',
                footer = footer, comments = ''
            )


    # --------------------------------------------------------------------------
    @staticmethod
    def write_mrc(path_mrc, data: "vg.Grid"):
        with gd.mrc.mrcfile.new(path_mrc, overwrite = True) as parser:
            parser.set_data(data.grid.astype(vg.FLOAT_DTYPE).transpose(2,1,0))
            parser.voxel_size = [data.dx, data.dy, data.dz]
            parser.header["origin"]['x'] = data.xmin # MRC convention
            parser.header["origin"]['y'] = data.ymin
            parser.header["origin"]['z'] = data.zmin
            parser.update_header_from_data()
            parser.update_header_stats()


    # --------------------------------------------------------------------------
    @staticmethod
    def write_ccp4(path_ccp4, data: "vg.Grid"):
        with gd.mrc.mrcfile.new(path_ccp4, overwrite = True) as parser:
            parser.set_data(data.grid.astype(vg.FLOAT_DTYPE).transpose(2,1,0))
            parser.voxel_size = [data.dx, data.dy, data.dz]
            parser.header["origin"]['x'] = data.xmin # MRC convention
            parser.header["origin"]['y'] = data.ymin
            parser.header["origin"]['z'] = data.zmin
            parser.header["nxstart"] = int(data.xmin / data.dx) # CCP4 convention
            parser.header["nystart"] = int(data.ymin / data.dy)
            parser.header["nzstart"] = int(data.zmin / data.dz)
            parser.update_header_from_data()
            parser.update_header_stats()


    # --------------------------------------------------------------------------
    @staticmethod
    def write_cmap(path_cmap, data: "vg.Grid", key):
        ### imitate the Chimera cmap format, as "specified" in this sample:
        ### https://github.com/RBVI/ChimeraX/blob/develop/testdata/cell15_timeseries.cmap
        def _add_generic_attrs(group, c = "GROUP"):
            group.attrs["CLASS"] = np.bytes_(c)
            group.attrs["TITLE"] = np.bytes_("")
            group.attrs["VERSION"] = np.bytes_("1.0")

        if not os.path.exists(path_cmap):
            with h5py.File(path_cmap, 'w') as h5:
                h5.attrs["PYTABLES_FORMAT_VERSION"] = np.bytes_("2.0")
                _add_generic_attrs(h5)

                chim = h5.create_group("Chimera")
                _add_generic_attrs(chim)

        with h5py.File(path_cmap, 'a') as parser:
            chim = parser["Chimera"]
            if key in chim.keys():
                frame = chim[key]
                if "data_zyx" in frame.keys():
                    del frame["data_zyx"]
            else:
                frame = parser.create_group(f"/Chimera/{key}")
                frame.attrs["chimera_map_version"] = np.int64(1)
                frame.attrs["chimera_version"] = np.bytes_(b'1.12_b40875')
                frame.attrs["name"] = np.bytes_(key)
                frame.attrs["origin"] = np.array([data.xmin, data.ymin, data.zmin], dtype = vg.FLOAT_DTYPE)
                frame.attrs["step"] = np.array([data.dz, data.dy, data.dx], dtype = vg.FLOAT_DTYPE)
                _add_generic_attrs(frame)

            framedata = frame.create_dataset(
                "data_zyx", data = data.grid.transpose(2,1,0), dtype = vg.FLOAT_DTYPE,
                compression = "gzip", compression_opts = vg.GZIP_COMPRESSION
            )
            _add_generic_attrs(framedata, "CARRAY")


    # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ OTHER I/O UTILITIES
    @staticmethod
    def read_auto(path_grid: Path) -> "vg.Grid":
        """Detect the format of the grid file based on its extension and then read it."""
        ext = path_grid.suffix.lower()

        # [TODO] improve the format detection?
        if ext == ".dx":
            return GridIO.read_dx(path_grid)

        if ext == ".mrc":
            return GridIO.read_mrc(path_grid)

        if ext == ".ccp4":
            return GridIO.read_ccp4(path_grid)

        if ext == ".cmap":
            keys = GridIO.get_cmap_keys(path_grid)
            if not keys: raise ValueError(f"Empty cmap file: {path_grid}")
            return GridIO.read_cmap(path_grid, keys[0])

        raise ValueError(f"Unrecognized file format: {ext}")


    # --------------------------------------------------------------------------
    @staticmethod
    def write_auto(path_grid: Path, data: "vg.Grid"):
        """Detect the format of the grid file based on its extension and then write it."""
        ext = path_grid.suffix.lower()

        # [TODO] improve the format detection?
        if ext == ".dx":
            GridIO.write_dx(path_grid, data)
            return

        if ext == ".mrc":
            GridIO.write_mrc(path_grid, data)
            return

        if ext == ".ccp4":
            GridIO.write_ccp4(path_grid, data)
            return

        if ext == ".cmap":
            keys = GridIO.get_cmap_keys(path_grid)
            GridIO.write_cmap(path_grid, data, key = f"map_{len(keys)+1}")
            return

        raise ValueError(f"Unrecognized file format: {ext}")


    # --------------------------------------------------------------------------
    @staticmethod
    def get_cmap_keys(path_cmap) -> list[str]:
        with h5py.File(path_cmap, 'r') as h5:
            return list(h5["Chimera"].keys())


# //////////////////////////////////////////////////////////////////////////////

# ------------------------------------------------------------------------------
def _read_mrc_ccp4(path_mrc, origin: np.ndarray) -> "vg.Grid":
    with gd.mrc.mrcfile.open(path_mrc) as parser:
        # machine_stamp = parser.header.machst
        ### [68 68 0 0] or [68 65 0 0] for little-endian <--- tested
        ### [17 17 0 0] for big-endian <--- what happens in these cases?

        vsize = np.array([
            parser.voxel_size['x'],
            parser.voxel_size['y'],
            parser.voxel_size['z'],
        ], dtype = vg.FLOAT_DTYPE)

        res = np.array([
            parser.header["mx"],
            parser.header["my"],
            parser.header["mz"],
        ], dtype = int)

        data: np.ndarray = parser.data.astype(vg.FLOAT_DTYPE)
        origin = origin.astype(vg.FLOAT_DTYPE)

        axes_correspondance =\
            parser.header.mapc, parser.header.mapr, parser.header.maps

        if axes_correspondance == (1, 2, 3):
            ms = vg.MolSystem.from_box_data(
                resolution = res.copy(), origin = origin.copy(), deltas = vsize.copy()
            )
            obj = vg.Grid(ms, init_grid = False)
            obj.grid = data.transpose(2,1,0)
            return obj

        if axes_correspondance == (3, 2, 1):
            ms = vg.MolSystem.from_box_data(
                resolution = res[::-1], origin = origin[::-1], deltas = vsize[::-1]
            )
            obj = vg.Grid(ms, init_grid = False)
            obj.grid = data
            return obj

        raise NotImplementedError(
            f"Unsupported axes correspondence in MRC file: {axes_correspondance}. "
            "Expected (1, 2, 3) or (3, 2, 1)."
        )


# ------------------------------------------------------------------------------



import numpy as np
import pandas as pd

import volgrids as vg

# //////////////////////////////////////////////////////////////////////////////
class GridVolumetricEnergy(vg.Grid):
    RADIUS_FIX = 0.4
    WIDTH_CYLINDERS = 0.25
    HEIGHT_DISKS = 0.25

    # --------------------------------------------------------------------------
    def __init__(self, ms: vg.MolSystem, df: pd.DataFrame, kind: str):
        super().__init__(ms)
        self.df = df[df["kind"] == kind].copy()
        self.kind = kind


    # --------------------------------------------------------------------------
    def populate_grid(self):
        for _,row in self.df.iterrows():
            if   row["energy"]  == 0: continue
            if   row["npoints"] == 2: self._process_2p_interaction(row)
            elif row["npoints"] == 3: self._process_3p_interaction(row)
            elif row["npoints"] == 4: self._process_4p_interaction(row)


    # --------------------------------------------------------------------------
    def _process_2p_interaction(self, row):
        a,b = self._get_positions(row)

        pos = (a + b) / 2  # kernel is placed at the center of the two particles
        direction = b - a  # get the direction vector between two particles

        ##### perform kernel operations
        radius = np.linalg.norm(direction) * self.RADIUS_FIX
        kernel = vg.KernelCylinder(
            radius = radius, vdirection = direction, width = self.WIDTH_CYLINDERS,
            deltas = self.ms.deltas, dtype = np.float32
        )
        self._apply_kernel(kernel, pos, row["energy"])


    # --------------------------------------------------------------------------
    def _process_3p_interaction(self, row):
        a, b, c = self._get_positions(row)
        u = vg.Math.normalize(a - b)
        v = vg.Math.normalize(c - b)
        p = vg.Math.normalize(b - a)
        q = vg.Math.normalize(c - a)

        pos = b                                     # kernel is placed at the vertix B of the triangle ABC
        direction = (u + v) / np.linalg.norm(u + v) # get the direction vector that bisects the angle between the two sides AB and BC
        angle = np.arccos(np.dot(u, v))             # get the angle between the two sides AB and BC
        normal = np.cross(p, q)                     # get the normal vector perpendicular to the plane ABC

        ##### perform kernel operations
        kernel = vg.KernelDiskConecut(
            radius = 2, vnormal = normal, height = self.HEIGHT_DISKS,
            vdirection = direction, max_angle = angle,
            deltas = self.ms.deltas, dtype = np.float32
        )
        self._apply_kernel(kernel, pos, row["energy"])


    # --------------------------------------------------------------------------
    def _process_4p_interaction(self, row):
        a,b,c,d = self._get_positions(row)

        pos = (a + b + c + d) / 4 # kernel is placed at the center of the four particles
        direction = a - c         # get the direction vector between the first  two non-adjacent particles
        normal = b - d            # get the direction vector between the second two non-adjacent particles

        ##### perform kernel operations
        radius0 = np.linalg.norm(direction) * self.RADIUS_FIX
        radius1 = np.linalg.norm(normal) * self.RADIUS_FIX
        kernel0 = vg.KernelCylinder(
            radius = radius0, vdirection = direction, width = self.WIDTH_CYLINDERS,
            deltas = self.ms.deltas, dtype = np.float32
        )
        kernel1 = vg.KernelCylinder(
            radius = radius1, vdirection = normal, width = self.WIDTH_CYLINDERS,
            deltas = self.ms.deltas, dtype = np.float32
        )
        self._apply_kernel(kernel0, pos, row["energy"])
        self._apply_kernel(kernel1, pos, row["energy"])


    # ------------------------------------------------------------------------------
    def _get_positions(self, row: pd.Series):
        ### idxs are expected to be 0-based
        def _split_idx_group(str_idxs: str) -> list[int]:
            return [int(idx) for idx in str_idxs.split('-')]

        if row["idxs_are_residues"]:
            return (
                self.ms.system.residues[idx].atoms.center_of_geometry()
                for idx in _split_idx_group(row["idxs"])
            )

        return self.ms.system.atoms[_split_idx_group(row["idxs"])].positions


    # --------------------------------------------------------------------------
    def _apply_kernel(self, kernel: vg.Kernel, position, energy):
        operation = "max" if energy > 0 else "min"
        kernel.link_to_grid(self.grid, self.ms.minCoords)
        kernel.stamp(position, multiplication_factor = energy, operation = operation)


# //////////////////////////////////////////////////////////////////////////////



import tempfile
import subprocess
from pathlib import Path
from MDAnalysis.core.groups import AtomGroup

import volgrids as vg

# //////////////////////////////////////////////////////////////////////////////
class APBSSubprocess:
    _PATH_SCRIPT = vg.resolve_path_package("utils/apbs.sh")

    # --------------------------------------------------------------------------
    def __init__(self, atoms: AtomGroup, name_pdb: str, keep_pqr: bool = False):
        self.atoms = atoms
        self.name_pdb = name_pdb
        self.keep_pqr = keep_pqr
        self._tmpdir = None

    # --------------------------------------------------------------------------
    def __enter__(self) -> Path:
        self._tmpdir = tempfile.TemporaryDirectory()
        path_tmpdir = Path(self._tmpdir.name)

        path_tmp_pdb  = path_tmpdir / self.name_pdb
        path_tmp_apbs = path_tmpdir / f"{self.name_pdb}.dx"

        self.atoms.write(path_tmp_pdb)
        proc = self.run_subprocess([str(path_tmp_pdb), "--pqr"])

        if proc.returncode != 0:
            self._tmpdir.cleanup()
            raise RuntimeError(f"apbs.sh failed (code={proc.returncode}):\n{proc.stderr}")
        if not path_tmp_apbs.exists():
            self._tmpdir.cleanup()
            raise FileNotFoundError(f"Expected APBS output not found: {path_tmp_apbs}")

        if self.keep_pqr:
            vg.PQR_CONTENTS_TEMP = (path_tmpdir / f"{self.name_pdb}.pqr").read_text()

        return path_tmp_apbs

    # --------------------------------------------------------------------------
    def __exit__(self, exc_type, exc, tb):
        if self._tmpdir is not None:
            self._tmpdir.cleanup()
        return False

    # --------------------------------------------------------------------------
    @classmethod
    def run_subprocess(self, args: list[str]) -> subprocess.CompletedProcess:
        return subprocess.run(
            ["/bin/bash", str(self._PATH_SCRIPT)] + args,
            capture_output = True, text = True
        )


# //////////////////////////////////////////////////////////////////////////////



from abc import ABC

import volgrids.smiffer as sm

# //////////////////////////////////////////////////////////////////////////////
class SmifHydro(sm.Smif, ABC):
    def iter_particles(self):
        for atom in self.ms.get_relevant_atoms():
            factor_res  = self.ms.chemtable.get_residue_hphob(atom)
            factor_atom = self.ms.chemtable.get_atom_hphob(atom)

            if (factor_res is None) and (factor_atom is None):
                continue # skip atoms with unknown name and resname

            if factor_res  is None: factor_res  = 1
            if factor_atom is None: factor_atom = 1

            yield atom, factor_res * factor_atom #/ len(atom.residue.atoms)


# //////////////////////////////////////////////////////////////////////////////



from abc import ABC, abstractmethod

import volgrids as vg
import volgrids.smiffer as sm

from .triplet import Triplet

# //////////////////////////////////////////////////////////////////////////////
class SmifHBonds(sm.Smif, ABC):
    # --------------------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.kernel: vg.KernelGaussianBivariateAngleDist = None
        self.hbond_getter: callable
        self.all_atoms = self.ms.get_relevant_atoms()
        self.res_atoms = None
        self.processed_interactors = set()

    # --------------------------------------------------------------------------
    @abstractmethod
    def find_tail_head_positions(self, triplet: Triplet) -> None:
        raise NotImplementedError()


    # --------------------------------------------------------------------------
    def populate_grid(self):
        for pos_interactor, vec_direction in self.iter_particles():
            self.kernel.recalculate_kernel(vec_direction, isStacking = False)
            self.kernel.stamp(pos_interactor, multiplication_factor = sm.ENERGY_SCALE)


    # --------------------------------------------------------------------------
    def iter_particles(self):
        for triplet in self._iter_triplets():
            self.find_tail_head_positions(triplet)
            vec_direction = triplet.get_direction_vector()

            if (triplet.pos_interactor is None) or (vec_direction is None):
                continue

            yield triplet.pos_interactor, vec_direction


    # --------------------------------------------------------------------------
    def _iter_triplets(self):
        for res in self.all_atoms.residues:
            hbond_tuples = self.hbond_getter(self.ms.chemtable, res.resname)
            if hbond_tuples is None: continue # skip weird residues

            self.processed_interactors.clear()

            for hbond_tuple in hbond_tuples:
                if not hbond_tuple: continue  # skip residues without HBond pairs

                triplet = Triplet(res, *hbond_tuple)
                self.res_atoms = self.all_atoms.select_atoms(triplet.str_this_res)
                triplet.set_pos_interactor(self.res_atoms)
                yield triplet


# //////////////////////////////////////////////////////////////////////////////



import numpy as np

import volgrids as vg
import volgrids.smiffer as sm

# //////////////////////////////////////////////////////////////////////////////
class Trimmer:
    KEY_INIT_COMMON_MASK = "mid" # the common mask is initialized by copying this specific mask

    def __init__(self, ms: "sm.MolSystemSmiffer", **distances):
        self.ms: "sm.MolSystemSmiffer" = ms

        self.distances: dict[str, float] = distances
        self.common_mask: vg.Grid = None
        self.specific_masks: dict[str, vg.Grid] = None


    # --------------------------------------------------------------------------
    @classmethod
    def init_infer_dists(cls, ms: "sm.MolSystemSmiffer") -> "sm.Trimmer":
        trimming_dists = {}
        if sm.DO_SMIF_HYDROPHILIC:
            trimming_dists["small"] = sm.TRIMMING_DIST_SMALL

        if (
            sm.DO_SMIF_STACKING or
            sm.DO_SMIF_HBA or sm.DO_SMIF_HBD or
            sm.DO_SMIF_HYDROPHOBIC or sm.SAVE_TRIMMING_MASK
        ):
            trimming_dists["mid"] = sm.TRIMMING_DIST_MID

        if sm.DO_SMIF_APBS:
            trimming_dists["large"] = sm.TRIMMING_DIST_LARGE

        return cls(ms, **trimming_dists)


    # --------------------------------------------------------------------------
    def reset_masks(self):
        self.common_mask: vg.Grid = None
        self.specific_masks = {k : vg.Grid(self.ms, dtype = bool) for k in self.distances.keys()}


    # --------------------------------------------------------------------------
    def trim(self):
        self.reset_masks()
        if sm.DO_TRIMMING_OCCUPANCY:
            self._trim_occupancies()

        if self._should_use_common_mask():
            self._init_common_mask()
            self._run_common_mask_operations()
            self._apply_common_mask_to_specific_masks()
            self._discard_common_mask()


    # --------------------------------------------------------------------------
    def mask_grid(self, smif: "vg.Grid", key: str):
        smif.grid[self.get_mask(key).grid] = 0


    # --------------------------------------------------------------------------
    def get_mask(self, key: str) -> vg.Grid:
        return self.specific_masks[key]


    # --------------------------------------------------------------------------
    def _should_use_common_mask(self) -> bool:
        return self.ms.do_ps


    # --------------------------------------------------------------------------
    def _init_common_mask(self):
        self.common_mask = self.specific_masks[self.KEY_INIT_COMMON_MASK].copy()


    # --------------------------------------------------------------------------
    def _run_common_mask_operations(self):
        if sm.DO_TRIMMING_FARAWAY:
            self._trim_faraway()

        if sm.DO_TRIMMING_SPHERE:
            self._trim_sphere()

        if sm.DO_TRIMMING_RNDS:
            self._trim_rnds()


    # --------------------------------------------------------------------------
    def _apply_common_mask_to_specific_masks(self):
        for mask in self.specific_masks.values():
            mask.grid |= self.common_mask.grid


    # --------------------------------------------------------------------------
    def _discard_common_mask(self):
        del self.common_mask
        self.common_mask = None


    # --------------------------------------------------------------------------
    def _trim_occupancies(self):
        for k,radius in self.distances.items():
            mask = self.specific_masks[k]
            kernel = vg.KernelSphere(radius, self.ms.deltas, bool)
            kernel.link_to_grid(mask.grid, self.ms.minCoords)

            for a in self.ms.get_relevant_atoms_broad(radius):
                kernel.stamp(a.position)


    # --------------------------------------------------------------------------
    def _trim_sphere(self):
        coords = vg.Math.get_coords_array(self.ms.resolution, self.ms.deltas, self.ms.minCoords)
        shifted_coords = coords - self.ms.cog
        dist_from_cog = vg.Math.get_norm(shifted_coords)
        self.common_mask.grid[dist_from_cog > self.ms.radius] = True


    # --------------------------------------------------------------------------
    def _trim_rnds(self):
        """
        Perform a random search to remove isolated regions.
        Can be problematic (e.g. slow, aggressive trimming); use with caution.
        """
        visited = np.zeros(self.ms.resolution, dtype = bool)

        directions = np.array([[i,j,k] for i in range(-1,2) for j in range(-1,2) for k in range(-1,2) if i&j&k])

        xres, yres, zres = self.ms.resolution
        xcog, ycog, zcog = np.floor(self.ms.resolution / 2).astype(int)
        cog_cube = set((x,y,z)
            for x in range(xcog - sm.COG_CUBE_RADIUS, xcog + sm.COG_CUBE_RADIUS + 1)
            for y in range(ycog - sm.COG_CUBE_RADIUS, ycog + sm.COG_CUBE_RADIUS + 1)
            for z in range(zcog - sm.COG_CUBE_RADIUS, zcog + sm.COG_CUBE_RADIUS + 1)
        )
        queue = cog_cube.copy()

        search_dist = np.full(self.ms.resolution, np.inf)
        for point in cog_cube: search_dist[point] = 0

        while queue:
            ### "random search" because popping from a set can be unpredictable
            i,j,k = node = queue.pop()
            visited[node] = True

            for dx,dy,dz in directions:
                ni = i + dx
                if not (0 <= ni < xres): continue

                nj = j + dy
                if not (0 <= nj < yres): continue

                nk = k + dz
                if not (0 <= nk < zres): continue

                neigh = ni,nj,nk
                search_dist[neigh] = min(search_dist[node] + 1, search_dist[neigh])
                if search_dist[neigh] > sm.MAX_RNDS_DIST: continue
                if visited[neigh]: continue
                if self.common_mask.grid[neigh]: continue

                queue.add(neigh)

        self.common_mask.grid[np.logical_not(visited)] = True


    # --------------------------------------------------------------------------
    def _trim_faraway(self):
        arr = np.zeros_like(self.common_mask.grid, dtype = bool)
        kernel = vg.KernelSphere(sm.TRIM_FARAWAY_DIST, self.ms.deltas, bool)
        kernel.link_to_grid(arr, self.ms.minCoords)
        for a in self.ms.get_relevant_atoms_broad(sm.TRIM_FARAWAY_DIST):
            kernel.stamp(a.position)

        self.common_mask.grid[~arr] = True


# //////////////////////////////////////////////////////////////////////////////



import volgrids as vg
import volgrids.smiffer as sm

from .hydro import SmifHydro

# //////////////////////////////////////////////////////////////////////////////
class SmifHydrophobic(SmifHydro):
    def populate_grid(self):
        radius = sm.MU_HYDROPHOBIC + sm.GAUSSIAN_KERNEL_SIGMAS * sm.SIGMA_HYDROPHOBIC
        kernel = vg.KernelGaussianUnivariateDist(
            radius, self.ms.deltas, vg.FLOAT_DTYPE, sm.PARAMS_HPHOB
        )
        kernel.link_to_grid(self.grid, self.ms.minCoords)

        for particle, mul_factor in self.iter_particles():
            if mul_factor < 0: continue
            kernel.stamp(particle.position, multiplication_factor =  mul_factor)


# //////////////////////////////////////////////////////////////////////////////



import numpy as np
from collections import defaultdict

import volgrids as vg
import volgrids.smiffer as sm

# //////////////////////////////////////////////////////////////////////////////
class SmifStacking(sm.Smif):
    def populate_grid(self):
        kernel = vg.KernelGaussianBivariateAngleDist(
            radius = sm.MU_DIST_STACKING + sm.GAUSSIAN_KERNEL_SIGMAS * sm.SIGMA_DIST_STACKING,
            deltas = self.ms.deltas, dtype = vg.FLOAT_DTYPE, params = sm.PARAMS_STACK
        )

        kernel.link_to_grid(self.grid, self.ms.minCoords)
        for res_atoms in self.iter_particles():
            cog = res_atoms.center_of_geometry()
            a,b,c = res_atoms.positions[:3]
            u = vg.Math.normalize(b - a)
            v = vg.Math.normalize(c - a)
            normal = vg.Math.normalize(np.cross(u, v))

            kernel.recalculate_kernel(normal, isStacking = True)
            kernel.stamp(cog, multiplication_factor = sm.ENERGY_SCALE)


    # --------------------------------------------------------------------------
    def iter_particles(self):
        resname_to_ids = defaultdict(set)
        atoms = self.ms.get_relevant_atoms()

        for a in atoms:
            resname_to_ids[a.resname.upper()].add((a.resid, a.chainID))

        for resname,res_infos in resname_to_ids.items():
            aromatic_atoms = self.ms.chemtable.get_names_stacking(resname)
            if aromatic_atoms is None: continue

            for resid,chain in res_infos:
                sel = f"resid {resid} and name {aromatic_atoms}"
                if chain: sel += f" and chainID {chain}"
                res_atoms = atoms.select_atoms(sel)
                if len(res_atoms) >= 3: # include rings even if they're not completely inside the grid's boundaries
                    yield res_atoms


# //////////////////////////////////////////////////////////////////////////////



from ._core.operations import VGOperations

from ._misc.comparison_result import ComparisonResult

from ._ui.param_handler import ParamHandlerVGTools
from ._ui.app import AppVGTools


############################# CONFIG FILE GLOBALS ##############################
__config_keys__ = set()


######################## COMMAND LINE ARGUMENTS GLOBALS ########################
### These are global variables that are to be set by
### an instance of ParamHandler (or its inherited classes)

OPERATION: str = '' # mode of the application, i.e. "convert", "pack", "unpack", "fix_cmap", "average", "compare"

import pathlib as _pathlib

### Convert
PATH_CONVERT_IN:   _pathlib.Path = None # "path/input/grid.dx" or "path/input/grid.mrc" or "path/input/grid.cmap"
PATH_CONVERT_DX:   _pathlib.Path = None # "path/input/grid.dx"
PATH_CONVERT_MRC:  _pathlib.Path = None # "path/output/grid.mrc"
PATH_CONVERT_CCP4: _pathlib.Path = None # "path/output/grid.ccp4"
PATH_CONVERT_CMAP: _pathlib.Path = None # "path/output/grid.cmap"

### Pack
PATHS_PACK_IN: list[_pathlib.Path] = None # list of paths to input grids for packing
PATH_PACK_OUT: _pathlib.Path = None # "path/output/packed.cmap"

### Unpack
PATH_UNPACK_IN:  _pathlib.Path = None # "path/input/packed.cmap"
PATH_UNPACK_OUT: _pathlib.Path = None # folder where to unpack the grids

### Fix CMAP
PATH_FIXCMAP_IN:  _pathlib.Path = None # "path/input/fix.cmap"
PATH_FIXCMAP_OUT: _pathlib.Path = None # "path/output/fix.cmap"

### Average
PATH_AVERAGE_IN:  _pathlib.Path = None # "path/input/traj.cmap"
PATH_AVERAGE_OUT: _pathlib.Path = None # "path/output/average.cmap"

### Compare
PATH_COMPARE_IN_0: _pathlib.Path = None # "path/input/grid_0.mrc"
PATH_COMPARE_IN_1: _pathlib.Path = None # "path/input/grid_1.mrc"
THRESHOLD_COMPARE: float # threshold for comparison (default 1e-5)



import numpy as np
from pathlib import Path

import volgrids as vg
import volgrids.vgtools as vgt

# //////////////////////////////////////////////////////////////////////////////
class VGOperations:
    @staticmethod
    def convert(path_in: Path, path_out: Path, fmt_out: vg.GridFormat):
        grid = vg.GridIO.read_auto(vgt.PATH_CONVERT_IN)

        func: callable = {
            vg.GridFormat.DX: vg.GridIO.write_dx,
            vg.GridFormat.MRC: vg.GridIO.write_mrc,
            vg.GridFormat.CCP4: vg.GridIO.write_ccp4,
            vg.GridFormat.CMAP: vg.GridIO.write_cmap,
        }.get(fmt_out, None)
        if func is None:
            raise ValueError(f"Unknown format for conversion: {fmt_out}")

        extra_args = (path_in.stem,) if fmt_out == vg.GridFormat.CMAP else ()
        func(path_out, grid, *extra_args)


    # --------------------------------------------------------------------------
    @staticmethod
    def pack(paths_in: list[Path], path_out: Path):
        resolution = None
        warned = False
        for path_in in paths_in:
            grid = vg.GridIO.read_auto(path_in)
            if resolution is None:
                resolution = f"{grid.xres} {grid.yres} {grid.zres}"

            new_res = f"{grid.xres} {grid.yres} {grid.zres}"
            if (new_res != resolution) and not warned:
                print(
                    f">>> Warning: Grid {path_in} has different resolution {new_res} than the first grid {resolution}. " +\
                    "Chimera won't recognize it as a volume series and open every grid in a separate representation. " +\
                    "Use `volgrids vgtools fix_cmap` if you want to fix this."
                )
                warned = True

            key = str(path_in.parent / path_in.stem).replace(' ', '_').replace('/', '_').replace('\\', '_')
            vg.GridIO.write_cmap(path_out, grid, key)


    # --------------------------------------------------------------------------
    @staticmethod
    def unpack(path_in: Path, folder_out: Path):
        keys = vg.GridIO.get_cmap_keys(path_in)
        for key in keys:
            path_out = folder_out / f"{key}.cmap"
            grid = vg.GridIO.read_cmap(path_in, key)
            vg.GridIO.write_cmap(path_out, grid, key)


    # --------------------------------------------------------------------------
    @staticmethod
    def fix_cmap(path_in: Path, path_out: Path):
        resolution = None
        keys = vg.GridIO.get_cmap_keys(path_in)
        for key in keys:
            grid = vg.GridIO.read_cmap(path_in, key)

            minCoords = (grid.xmin, grid.ymin, grid.zmin)
            maxCoords = (grid.xmax, grid.ymax, grid.zmax)
            if resolution is None:
                resolution = (grid.xres, grid.yres, grid.zres)

            grid.reshape(minCoords, maxCoords, resolution)
            vg.GridIO.write_cmap(path_out, grid, key)


    # --------------------------------------------------------------------------
    @staticmethod
    def average(path_in: Path, path_out: Path):
        keys = vg.GridIO.get_cmap_keys(path_in)
        nframes = len(keys)

        grid = vg.GridIO.read_cmap(path_in, keys[0])
        avg = np.zeros_like(grid.grid)
        for key in keys:
            print(key)
            avg += vg.GridIO.read_cmap(path_in, key).grid
        avg /= nframes

        grid_avg: vg.Grid = vg.Grid(grid.ms, init_grid = False)
        grid_avg.grid = avg

        vg.GridIO.write_auto(path_out, grid_avg)


    # --------------------------------------------------------------------------
    @staticmethod
    def compare(path_in_0: Path, path_in_1: Path, threshold: float) -> "vgt.ComparisonResult":
        def _are_different_vector(vec0, vec1):
            diff = np.abs(vec0 - vec1)
            return len(diff[diff > threshold]) != 0

        grid_0 = vg.GridIO.read_auto(path_in_0)
        grid_1 = vg.GridIO.read_auto(path_in_1)

        deltas_0     = grid_0.get_deltas();     deltas_1     = grid_1.get_deltas()
        resolution_0 = grid_0.get_resolution(); resolution_1 = grid_1.get_resolution()
        min_coords_0 = grid_0.get_min_coords(); min_coords_1 = grid_1.get_min_coords()
        max_coords_0 = grid_0.get_max_coords(); max_coords_1 = grid_1.get_max_coords()

        if _are_different_vector(resolution_0, resolution_1):
            return vgt.ComparisonResult(0, 0, 0.0, 0.0,
                [f"Warning: Grids {path_in_0} and {path_in_1} have different shapes: {resolution_0} vs {resolution_1}. Aborting."]
            )

        warnings = []
        if _are_different_vector(min_coords_0, min_coords_1):
            warnings.append(
                f"Warning: Grids {path_in_0} and {path_in_1} have different origin: {min_coords_0} vs {min_coords_1}. Comparison may not be accurate."
            )
        if _are_different_vector(max_coords_0, max_coords_1):
            warnings.append(
                f"Warning: Grids {path_in_0} and {path_in_1} have different max coordinate: {max_coords_0} vs {max_coords_1}. Comparison may not be accurate."
            )
        if _are_different_vector(deltas_0, deltas_1):
            warnings.append(
                f"Warning: Grids {path_in_0} and {path_in_1} have different deltas: {deltas_0} vs {deltas_1}. Comparison may not be accurate."
            )

        diff = abs(grid_1 - grid_0)
        mask = diff.grid > threshold

        npoints_diff  = len(mask[mask])
        npoints_total = grid_0.xres * grid_0.yres * grid_0.zres
        cumulative_diff = np.sum(diff.grid[mask])
        avg_diff = (cumulative_diff / npoints_diff) if (npoints_diff > 0) else 0

        return vgt.ComparisonResult(npoints_diff, npoints_total, cumulative_diff, avg_diff, warnings)



# //////////////////////////////////////////////////////////////////////////////



import volgrids as vg
import volgrids.vgtools as vgt

# //////////////////////////////////////////////////////////////////////////////
class AppVGTools(vg.App):
    CONFIG_MODULES = (vg,)
    _CLASS_PARAM_HANDLER = vgt.ParamHandlerVGTools

    # --------------------------------------------------------------------------
    def run(self) -> None:
        if vgt.OPERATION == "convert":
            self._run_convert()
            return

        if vgt.OPERATION == "pack":
            print(f">>> Packing {len(vgt.PATHS_PACK_IN)} grids into '{vgt.PATH_PACK_OUT}'")
            vgt.VGOperations.pack(vgt.PATHS_PACK_IN, vgt.PATH_PACK_OUT)
            return

        if vgt.OPERATION == "unpack":
            print(f">>> Unpacking '{vgt.PATH_UNPACK_IN}' into '{vgt.PATH_UNPACK_OUT}'")
            vgt.VGOperations.unpack(vgt.PATH_UNPACK_IN, vgt.PATH_UNPACK_OUT)
            return

        if vgt.OPERATION == "fix_cmap":
            print(f">>> Fixing CMAP file: {vgt.PATH_FIXCMAP_IN}")
            vgt.VGOperations.fix_cmap(vgt.PATH_FIXCMAP_IN, vgt.PATH_FIXCMAP_OUT)
            return

        if vgt.OPERATION == "average":
            print(f">>> Averaging CMAP file: {vgt.PATH_AVERAGE_IN}")
            vgt.VGOperations.average(vgt.PATH_AVERAGE_IN, vgt.PATH_AVERAGE_OUT)
            return

        if vgt.OPERATION == "compare":
            print(f">>> Comparing grids: {vgt.PATH_COMPARE_IN_0} vs {vgt.PATH_COMPARE_IN_1} (threshold={vgt.THRESHOLD_COMPARE:2.2e})")
            result = vgt.VGOperations.compare(vgt.PATH_COMPARE_IN_0, vgt.PATH_COMPARE_IN_1, vgt.THRESHOLD_COMPARE)

            for message in result.messages:
                print(f"...>>> {message}")
            if result.npoints_total == 0: return

            print(
                f"...>>> {result.npoints_diff}/{result.npoints_total} points differ " +\
                f"({100 * result.npoints_diff / result.npoints_total:.2f}%)\n" +\
                f"...>>> Accumulated difference: {result.cumulative_diff:2.2e} " +\
                f"(avg {result.avg_diff:2.2e} per point)"
            )
            return

        raise ValueError(f"Unknown mode: {vgt.OPERATION}")


    # --------------------------------------------------------------------------
    def _import_config_dependencies(self):
        import numpy as np
        return {"np": np, "vg": vg}


    # --------------------------------------------------------------------------
    def _run_convert(self):
        def _convert(path_out, fmt_out: vg.GridFormat):
            if path_out is None: return
            print(f">>> Converting {vgt.PATH_CONVERT_IN} file to {fmt_out.name}: {path_out}")
            vgt.VGOperations.convert(vgt.PATH_CONVERT_IN, path_out, fmt_out)

        _convert(vgt.PATH_CONVERT_DX,   vg.GridFormat.DX)
        _convert(vgt.PATH_CONVERT_MRC,  vg.GridFormat.MRC)
        _convert(vgt.PATH_CONVERT_CCP4, vg.GridFormat.CCP4)
        _convert(vgt.PATH_CONVERT_CMAP, vg.GridFormat.CMAP)


# //////////////////////////////////////////////////////////////////////////////



import volgrids as vg
import volgrids.veins as ve

# //////////////////////////////////////////////////////////////////////////////
class ParamHandlerVeins(vg.ParamHandler):
    _EXPECTED_CLI_FLAGS = {
        "help"  : ("-h", "--help"),
        "output": ("-o", "--output"),
        "traj"  : ("-t", "--trajectory"),
        "cutoff": ("-c", "--cutoff"),
    }
    _DEFAULT_ENERGY_CUTOFF = 1e-3


    # --------------------------------------------------------------------------
    def assign_globals(self):
        self._set_help_str(
            "[WIP] WORK IN PROGRESS - NOT YET FUNCTIONAL",
            "usage: volgrids veins [mode] [options...]",
            "Available modes:",
            "  energies - Generate grids to visually represent in space the interaction energies of a molecular system.",
            # "  force  - ", # TODO: Implement force mode
            "Run 'volgrids veins [mode] --help' for more details on each mode.",
        )
        if self._has_param_kwds("help") and not self._has_params_pos():
            self._exit_with_help()

        ve.MODE = self._safe_get_param_pos(0).lower()
        func: callable = self._safe_map_value(ve.MODE,
            energies = self._parse_energies,
            forces   = self._parse_forces,
        )
        func()


    # --------------------------------------------------------------------------
    def _parse_energies(self) -> None:
        self._set_help_str(
            "usage: volgrids veins energy [path/input/structure.pdb] [path/input/energies.csv] [options...]",
            "Available options:",
            "-h, --help       Show this help message and exit.",
            "-o, --output     Path to the folder where the output SMIFs should be stored. If not provided, the parent folder of the input structure file will be used.",
            "-t, --trajectory Path to a trajectory file (e.g. XTC) supported by MDAnalysis. In this case, the energies CSV file contains an energy column for each frame. The header of such columns must start with 'frame'.",
            "-c, --cutoff     Energies below this cutoff will be ignored. Default value: 1e-3.",
        )
        if self._has_param_kwds("help"):
            self._exit_with_help()

        ve.PATH_STRUCT = self._safe_path_file_in(
            self._safe_get_param_pos(1,
               err_msg = "No input structure file provided. Provide a path to the structure file as first positional argument."
            )
        )

        ve.PATH_ENERGIES_CSV = self._safe_path_file_in(
            self._safe_get_param_pos(2,
               err_msg = "No energies CSV file provided. Provide a path to the energies file as second positional argument."
            )
        )

        ve.FOLDER_OUT = self._safe_kwd_folder_out("output", default = ve.PATH_STRUCT.parent)
        ve.PATH_TRAJ  = self._safe_kwd_file_in("traj")
        ve.ENERGY_CUTOFF = self._safe_kwd_float("cutoff", default = self._DEFAULT_ENERGY_CUTOFF)


    # --------------------------------------------------------------------------
    def _parse_forces(self) -> None:
        raise NotImplementedError("VEINS Forces mode is not yet implemented.")


# //////////////////////////////////////////////////////////////////////////////



import volgrids as vg
import volgrids.vgtools as vgt

# //////////////////////////////////////////////////////////////////////////////
class ParamHandlerVGTools(vg.ParamHandler):
    _EXPECTED_CLI_FLAGS = {
            "help"   : ("-h", "--help"),
            "input"  : ("-i", "--input"),
            "output" : ("-o", "--output"),
            "dx"     : ("-d", "--dx"),
            "mrc"    : ("-m", "--mrc"),
            "ccp4"   : ("-p", "--ccp4"),
            "cmap"   : ("-c", "--cmap"),
            "thresh" : ("-t", "--threshold"),
    }
    _DEFAULT_COMPARISON_THRESHOLD = 1e-5


    # --------------------------------------------------------------------------
    def assign_globals(self):
        self._set_help_str(
            "usage: volgrids vgtools [convert|pack|unpack|average|fix_cmap] [options...]",
            "Available modes:",
            "  convert  - Convert grid files between formats.",
            "  pack     - Pack multiple grid files into a single CMAP series-file.",
            "  unpack   - Unpack a CMAP series-file into multiple grid files.",
            "  average  - Average all grids in a CMAP series-file into a single grid.",
            "  fix_cmap - Ensure that all grids in a CMAP series-file have the same resolution, interpolating them if necessary.",
            "  compare  - Compare two grid files by printing the number of differing points and their accumulated difference.",
            "Run 'volgrids vgtools [mode] --help' for more details on each mode.",
        )
        if self._has_param_kwds("help") and not self._has_params_pos():
            self._exit_with_help()

        vgt.OPERATION = self._safe_get_param_pos(0).lower()
        func: callable = self._safe_map_value(vgt.OPERATION,
            convert  = self._parse_convert,
            pack     = self._parse_pack,
            unpack   = self._parse_unpack,
            fix_cmap = self._parse_fix_cmap,
            average  = self._parse_average,
            compare  = self._parse_compare,
        )
        func()


    # --------------------------------------------------------------------------
    def _parse_convert(self) -> None:
        self._set_help_str(
            "usage: volgrids vgtools convert [path/input/grid] [options...]",
            "Available options:",
            "-h, --help  Show this help message and exit.",
            "-d, --dx    File path where to save the converted grid in DX format.",
            "-m, --mrc   File path where to save the converted grid in MRC format.",
            "-p, --ccp4  File path where to save the converted grid in CCP4 format.",
            "-c, --cmap  File path where to save the converted grid in CMAP format. The stem of the input file will be used as the CMAP key.",
        )
        if self._has_param_kwds("help"):
            self._exit_with_help()

        vgt.PATH_CONVERT_IN = self._safe_path_file_in(
            self._safe_get_param_pos(1,
               err_msg = "No input grid file provided. Provide a path to the grid file as second positional argument."
            )
        )

        vgt.PATH_CONVERT_DX   = self._safe_kwd_file_out("dx")
        vgt.PATH_CONVERT_MRC  = self._safe_kwd_file_out("mrc")
        vgt.PATH_CONVERT_CCP4 = self._safe_kwd_file_out("ccp4")
        vgt.PATH_CONVERT_CMAP = self._safe_kwd_file_out("cmap")


    # --------------------------------------------------------------------------
    def _parse_pack(self) -> None:
        self._set_help_str(
            "usage: volgrids vgtools pack [options...]",
            "Available options:",
            "-h, --help    Show this help message and exit.",
            "-i, --input   List of file paths with the input grids to be packed. At least one grid file must be provided.",
            "-o, --output  File path where to save the packed grid in CMAP format. Must be provided.",
        )
        if self._has_param_kwds("help"):
            self._exit_with_help()

        vgt.PATHS_PACK_IN = [
            self._safe_path_file_in(path) for path in \
            self._safe_get_param_kwd_list("input")
        ]

        vgt.PATH_PACK_OUT = self._safe_kwd_file_out("output", required = True)


    # --------------------------------------------------------------------------
    def _parse_unpack(self) -> None:
        self._set_help_str(
            "usage: volgrids vgtools unpack [options...]",
            "Available options:",
            "-h, --help    Show this help message and exit.",
            "-i, --input   File path to the CMAP series-file to be unpacked. Must be provided.",
            "-o, --output  Folder path where to save the unpacked grids. If not provided, the parent folder of the input packed file will be used.",
        )
        if self._has_param_kwds("help"):
            self._exit_with_help()

        vgt.PATH_UNPACK_IN  = self._safe_kwd_file_in("input", required = True)
        vgt.PATH_UNPACK_OUT = self._safe_kwd_folder_out("output", default = vgt.PATH_UNPACK_IN.parent)


    # --------------------------------------------------------------------------
    def _parse_fix_cmap(self) -> None:
        self._set_help_str(
            "usage: volgrids vgtools fix_cmap [options...]",
            "Available options:",
            "-h, --help    Show this help message and exit.",
            "-i, --input   File path to the CMAP file to be fixed. Must be provided.",
            "-o, --output  File path where to save the fixed CMAP file. Must be provided.",
        )
        if self._has_param_kwds("help"):
            self._exit_with_help()

        vgt.PATH_FIXCMAP_IN  = self._safe_kwd_file_in("input", required = True)
        vgt.PATH_FIXCMAP_OUT = self._safe_kwd_file_out("output", required = True)


    # --------------------------------------------------------------------------
    def _parse_average(self) -> None:
        self._set_help_str(
            "usage: volgrids vgtools average [options...]",
            "Available options:",
            "-h, --help    Show this help message and exit.",
            "-i, --input   File path to the CMAP series-file to be averaged. Must be provided.",
            "-o, --output  File path where to save the averaged CMAP file. Must be provided.",
        )
        if self._has_param_kwds("help"):
            self._exit_with_help()

        vgt.PATH_AVERAGE_IN  = self._safe_kwd_file_in("input", required = True)
        vgt.PATH_AVERAGE_OUT = self._safe_kwd_file_out("output", required = True)


    # --------------------------------------------------------------------------
    def _parse_compare(self) -> None:
        self._set_help_str(
            "usage: volgrids vgtools compare [path/input/grid_0] [path/input/grid_1] [options...]",
            "Available options:",
            "-h, --help       Show this help message and exit.",
            "-t, --threshold  Threshold for comparison. Default is 1e-3.",
        )
        if self._has_param_kwds("help"):
            self._exit_with_help()

        vgt.PATH_COMPARE_IN_0 = self._safe_path_file_in(
            self._safe_get_param_pos(1,
               err_msg = "No first input grid file provided. Provide a path to the first grid file as second positional argument."
            )
        )

        vgt.PATH_COMPARE_IN_1 = self._safe_path_file_in(
            self._safe_get_param_pos(2,
               err_msg = "No second input grid file provided. Provide a path to the second grid file as third positional argument."
            )
        )

        vgt.THRESHOLD_COMPARE = self._safe_kwd_float("thresh", default = self._DEFAULT_COMPARISON_THRESHOLD)


# //////////////////////////////////////////////////////////////////////////////



import numpy as np

import volgrids as vg
import volgrids.smiffer as sm

# //////////////////////////////////////////////////////////////////////////////
class SmifAPBS(sm.Smif):
    # --------------------------------------------------------------------------
    def populate_grid(self):
        if sm.PATH_APBS is not None:
            self.apbs_to_smif(sm.PATH_APBS)
            return

        timer = vg.Timer().start()
        with vg.APBSSubprocess(
            self.ms.system.atoms, sm.PATH_STRUCT.name, keep_pqr = True
        ) as path_apbs: self.apbs_to_smif(path_apbs, timer)


    # --------------------------------------------------------------------------
    def apbs_to_smif(self, path_apbs_in, timer: vg.Timer = None):
        if timer is not None:
            sm.APBS_ELAPSED_TIME = timer.end(text = "APBS", end = ' ')

        apbs = vg.GridIO.read_auto(path_apbs_in)
        apbs.reshape_as_other(self)
        self.grid = apbs.grid


    # --------------------------------------------------------------------------
    def apply_logabs_transform(self):
        if self.is_empty():
            print(f"...--- APBS potential grid is empty. Skipping logabs transform.", flush = True)
            return

        logpos = np.log10( self.grid[self.grid > 0])
        logneg = np.log10(-self.grid[self.grid < 0])

        ##### APPLY CUTOFFS
        logpos[logpos < sm.APBS_MIN_CUTOFF] = sm.APBS_MIN_CUTOFF
        logneg[logneg < sm.APBS_MIN_CUTOFF] = sm.APBS_MIN_CUTOFF
        logpos[logpos > sm.APBS_MAX_CUTOFF] = sm.APBS_MAX_CUTOFF
        logneg[logneg > sm.APBS_MAX_CUTOFF] = sm.APBS_MAX_CUTOFF

        ##### SHIFT VALUES TO 0
        logpos -= sm.APBS_MIN_CUTOFF
        logneg -= sm.APBS_MIN_CUTOFF

        ##### REVERSE SIGN OF LOG(ABS(GRID_NEG)) AND DOUBLE BOTH
        logpos *=  2 # this way the range of points varies between
        logneg *= -2 # 2*APBS_MIN_CUTOFF and 2*APBS_MAX_CUTOFF

        ##### RESULT
        self.grid[self.grid > 0] = logpos
        self.grid[self.grid < 0] = logneg


# //////////////////////////////////////////////////////////////////////////////



import sys
import warnings
from pathlib import Path

try:
    import volgrids as vg
except ImportError:
    sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
    import volgrids as vg

# ------------------------------------------------------------------------------
def help_and_exit(exit_code: int):
    print(
        f"usage: volgrids [smiffer|veins|vgtools] [options...]",
         "Available applications:",
         "  apbs     - Generate raw APBS potential grids for biomolecular structures.",
         "  smiffer  - Calculate SMIFs for biomolecular structures.",
         "  veins    - Calculate VEINS for biomolecular structures.",
         "  vgtools  - Miscellaneous tools for volumetric grids.",
        f"Run 'volgrids [app] --help' for more details on each application.",
        sep = '\n'
    )
    exit(exit_code)


# ------------------------------------------------------------------------------
def path_default_config() -> Path | None:
    if not (__package__ == '' or __package__ is None):
        return None
    path_config = Path(__file__).parent.parent / "config_volgrids.ini"
    return path_config if path_config.is_file() else None


# ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def main():
    vg.PATH_DEFAULT_CONFIG = path_default_config()

    argv = sys.argv[1:]
    if not argv: help_and_exit(1)
    if argv[0] in ("-h", "--help"): help_and_exit(0)

    warnings.filterwarnings("ignore", module = "MDAnalysis.*")
    app_name = argv[0].lower()
    app_args = argv[1:]

    if app_name == "apbs":
        print(f">>> Launching APBS subprocess for '{app_args[0]}'...", flush = True)
        apbs = vg.APBSSubprocess.run_subprocess(app_args)
        print(f"{apbs.stdout}\n{apbs.stderr}".strip(), flush = True)
        exit(apbs.returncode)

    if app_name == "smiffer":
        import volgrids.smiffer as sm
        sm.AppSmiffer.from_cli(app_args).run()
        exit(0)

    if app_name == "veins":
        import volgrids.veins as ve
        ve.AppVeins.from_cli(app_args).run()
        exit(0)

    if app_name == "vgtools":
        import volgrids.vgtools as vgt
        vgt.AppVGTools.from_cli(app_args).run()
        exit(0)

    help_and_exit(2)


################################################################################
if __name__ == "__main__":
    main()


################################################################################



import pandas as pd

import volgrids as vg
import volgrids.veins as ve

# ------------------------------------------------------------------------------
def _assert_df(df: pd.DataFrame, *cols_metadata):
    if not set(df.columns).issuperset(cols_metadata):
        raise ValueError(
            f"CSV file '{ve.PATH_ENERGIES_CSV}' must contain the columns: " +\
            ", ".join(map(lambda x: f"'{x}'", cols_metadata)) + " " +\
            f"Found columns: {df.columns}"
        )


# //////////////////////////////////////////////////////////////////////////////
class AppVeins(vg.App):
    CONFIG_MODULES = (vg,)
    _CLASS_PARAM_HANDLER = ve.ParamHandlerVeins

    # --------------------------------------------------------------------------
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.ms = vg.MolSystem(ve.PATH_STRUCT, ve.PATH_TRAJ)
        self.df = pd.read_csv(ve.PATH_ENERGIES_CSV).dropna(how = "any")
        self.cols_frames: list = None

        if self.ms.do_traj:
            _assert_df(self.df, "kind", "npoints", "idxs", "idxs_are_residues")
            self.cols_frames = sorted(filter(lambda x: x.startswith("frame"), self.df.columns))
            if not self.cols_frames:
                raise ValueError(
                    f"CSV file '{ve.PATH_ENERGIES_CSV}' must contain at least one column starting with 'frame' "
                    "when running in trajectory mode."
                )
            mat = self.df[self.cols_frames].to_numpy()
            mat[mat < ve.ENERGY_CUTOFF] = 0.0
            self.df.loc[:, self.cols_frames] = mat

        else:
            _assert_df(self.df, "kind", "npoints", "idxs", "idxs_are_residues", "energy")
            self.df = self.df[self.df["energy"].abs() > ve.ENERGY_CUTOFF]

        self.timer = vg.Timer(
            f">>> Now processing '{self.ms.molname}' ({ve.MODE})"
        )


    # --------------------------------------------------------------------------
    def run(self):
        self.timer.start()

        if self.ms.do_traj: # TRAJECTORY MODE
            for _ in self.ms.system.trajectory:
                current_col = self.cols_frames[self.ms.frame]
                self.df["energy"] = self.df[current_col]
                self.ms.frame += 1

                timer_frame = vg.Timer(f"...>>> Frame {self.ms.frame}/{len(self.ms.system.trajectory)}")
                timer_frame.start()
                self._process_grids()
                timer_frame.end()

        else: # SINGLE PDB MODE
            self._process_grids()

        self.timer.end()


    # --------------------------------------------------------------------------
    def _import_config_dependencies(self):
        import numpy as np
        return {"np": np, "vg": vg}


    # --------------------------------------------------------------------------
    def _process_grids(self):
        for kind in self.df["kind"].unique():
            grid = ve.GridVolumetricEnergy(self.ms, self.df, kind)
            grid.populate_grid()
            grid.save_data(ve.FOLDER_OUT, grid.kind)


# # //////////////////////////////////////////////////////////////////////////////



import numpy as np

import volgrids as vg

# ------------------------------------------------------------------------------
def _clamp_indices(g_idx0, g_idx1, k_idx0, k_idx1, g_res, k_res):
    if (g_idx0 < 0) and (g_res <= g_idx1): # kernel would be larger than the big grid
        return 0, g_res, -g_idx0, -g_idx0+g_res

    if g_idx0 < 0: # kernel would start before the big grid
        return 0, g_idx1, k_res-g_idx1, k_idx1

    if g_idx1 >= g_res: # kernel would end after the big grid
        return g_idx0, g_res, k_idx0, g_res-g_idx0

    return g_idx0, g_idx1, k_idx0, k_idx1


# //////////////////////////////////////////////////////////////////////////////
class Kernel:
    def __init__(self, radius, deltas, dtype, operation = "sum"):
        ##### store kernel values
        self.kernel_res = (np.ceil(radius / deltas) * 2 + 1).astype(int)
        self.deltas = deltas
        self.kernel = np.zeros(self.kernel_res, dtype = dtype)

        ##### initialize empty big-grid values (assign them later with link_to_grid)
        self.grid = None
        self.grid_origin = None
        self.grid_res = None

        ##### initizalize auxiliary kernel of distance values
        self.center = np.floor(self.kernel_res / 2) * self.deltas
        self.coords = vg.Math.get_coords_array(self.kernel_res, self.deltas)
        self.shifted_coords = self.coords - self.center
        self.dist = vg.Math.get_norm(self.shifted_coords)

        ##### set operation
        self.operation: callable[np.array, np.array]
        if   operation == "sum": self.operation = np.add
        elif operation == "min": self.operation = np.minimum
        elif operation == "max": self.operation = np.maximum
        else: raise ValueError(f"Unknown operation: {operation}. Use 'sum', 'min' or 'max'.")


    # --------------------------------------------------------------------------
    def link_to_grid(self, grid, grid_origin):
        self.grid = grid
        self.grid_origin = grid_origin
        self.grid_res = np.array(grid.shape)


    # --------------------------------------------------------------------------
    def stamp(self, center_stamp_at, multiplication_factor = None, operation = "sum"):
        if self.grid is None:
            raise ValueError("No grid associated, can't stamp Kernel. Use 'link_to_grid' first.")

        ##### infer the position where to stamp the kernel at the big grid
        stamp_orig = center_stamp_at - self.deltas * self.kernel_res / 2
        rel_orig = stamp_orig - self.grid_origin
        idx_start = np.round(rel_orig / self.deltas).astype(int)
        idx_end = idx_start + self.kernel_res

        ##### skip cases where the kernel would be stamped outside the big grid
        if (idx_end < 0).any(): return
        if (idx_start > self.grid_res).any(): return

        ##### initialize the grid (g_*) and kernel (k_*) indices
        g_i0, g_j0, g_k0 = idx_start
        g_i1, g_j1, g_k1 = idx_end
        k_i0, k_j0, k_k0 = 0, 0, 0
        k_i1, k_j1, k_k1 = self.kernel_res

        g_rx, g_ry, g_rz = self.grid_res
        k_rx, k_ry, k_rz = self.kernel_res

        ##### clamp the indices of both the big grid and the kernel
        g_i0, g_i1, k_i0, k_i1 = _clamp_indices(g_i0, g_i1, k_i0, k_i1, g_rx, k_rx)
        g_j0, g_j1, k_j0, k_j1 = _clamp_indices(g_j0, g_j1, k_j0, k_j1, g_ry, k_ry)
        g_k0, g_k1, k_k0, k_k1 = _clamp_indices(g_k0, g_k1, k_k0, k_k1, g_rz, k_rz)

        ##### stamp the kernel on the big grid
        subkernel = self.kernel[k_i0:k_i1, k_j0:k_j1, k_k0:k_k1]
        subgrid   = self.grid  [g_i0:g_i1, g_j0:g_j1, g_k0:g_k1]

        ### multiplication_factor defaults to None (instead of 1) to avoid problems with bool grids
        scaled_subkernel = subkernel if (multiplication_factor is None) else multiplication_factor * subkernel

        self.grid[g_i0:g_i1, g_j0:g_j1, g_k0:g_k1] = self.operation(subgrid, scaled_subkernel)


# //////////////////////////////////////////////////////////////////////////////



from collections import defaultdict

import volgrids as vg

# ------------------------------------------------------------------------------
def _parse_atoms_triplet(triplet: str) -> tuple[str, str, str, str, bool]:
    def _assert(condition: bool):
        assert condition, \
            f"Triplet '{triplet}' is not in the expected formats 'I=T->H' or 'I=T0.T1->H'."

    stripped = triplet.strip('!')
    hbond_fixed = stripped != triplet

    parts = stripped.split('=')
    _assert(len(parts) == 2)

    interactor, direction = parts
    parts = direction.split("->")
    _assert(len(parts) == 2)

    tail, head = parts
    _assert(tail and head and interactor)

    ### valid syntax?  | yes | no |
    ### 'i=t->h'       |  X  |    |
    ### 'i=t0.t1->h'   |  X  |    |
    ### 'i=t0.t1.t2->h'|  X  |    |
    ### 'i=t0.->h'     |     | X  |
    ### 'i=.t1->h'     |     | X  |
    tail_points = tail.split('.')
    _assert(len(tail_points) > 0 and all(tail_points))

    return interactor, tail_points, head, hbond_fixed


# //////////////////////////////////////////////////////////////////////////////
class ParserChemTable:
    def __init__(self, path_table):
        self.selection_query: str = ''
        self._parser_ini = vg.ParserIni(path_table)
        self._residues_hphob: dict[str, float] = {}
        self._atoms_hphob: defaultdict[str, dict[str, float]] = defaultdict(dict)
        self._names_stk: dict[str, str] = {}
        self._names_hba: dict[str, list[tuple[str, str, str, bool]]] = {}
        self._names_hbd: dict[str, list[tuple[str, str, str, bool]]] = {}
        self._parse_table()


    # --------------------------------------------------------------------------
    def get_residue_hphob(self, atom):
        return self._residues_hphob.get(atom.resname)


    # --------------------------------------------------------------------------
    def get_atom_hphob(self, atom):
        dict_resid = self._atoms_hphob.get(atom.resname)
        if dict_resid is None: return None
        return dict_resid.get(atom.name)


    # --------------------------------------------------------------------------
    def get_names_stacking(self, resname: str):
        return self._names_stk.get(resname)


    # --------------------------------------------------------------------------
    def get_names_hba(self, resname: str):
        return self._names_hba.get(resname)


    # --------------------------------------------------------------------------
    def get_names_hbd(self, resname: str):
        return self._names_hbd.get(resname)


    # --------------------------------------------------------------------------
    def _parse_table(self):
        ### extract values from the lines
        query = self._parser_ini.get("SELECTION_QUERY")
        if query is None: raise ValueError("No selection query found in the table file.")
        self.selection_query = query[0]

        for resname, value in self._parser_ini.iter_splitted_lines("RES_HPHOBICITY", sep = ':'):
            self._residues_hphob[resname] = float(value)

        for names, value in self._parser_ini.iter_splitted_lines("ATOM_HPHOBICITY", sep = ':'):
            resname, atomname = names.split('/')
            self._atoms_hphob[resname][atomname] = float(value)

        for resname, atomnames in self._parser_ini.iter_splitted_lines("NAMES_STACKING", sep = ':'):
            self._names_stk[resname] = atomnames

        for resname, str_triplets in self._parser_ini.iter_splitted_lines("NAMES_HBACCEPTORS", sep = ':'):
            triplets = map(_parse_atoms_triplet, str_triplets.split())
            self._names_hba[resname] = [(hba,tail,head,False) for hba,tail,head,_ in triplets] # hbond_fixed must always be False for HBAcceptors

        for resname, str_triplets in self._parser_ini.iter_splitted_lines("NAMES_HBDONORS", sep = ':'):
            triplets = list(map(_parse_atoms_triplet, str_triplets.split()))
            self._names_hbd[resname] = triplets

        ### expand the atoms declared with the '*' wildcard to all residues (hydrophobicity)
        hphob_wildcard = self._atoms_hphob.get('*')
        if hphob_wildcard is None: return

        del self._atoms_hphob['*']
        for res_dict in self._atoms_hphob.values():
            res_dict.update(hphob_wildcard)


# //////////////////////////////////////////////////////////////////////////////



import numpy as np
import MDAnalysis as mda
from MDAnalysis.core.groups import Residue

import volgrids as vg

# ------------------------------------------------------------------------------
def _safe_return_coords(atoms: mda.AtomGroup, sel_string: str):
    sel_atoms = atoms.select_atoms(sel_string)
    if len(sel_atoms) == 0: return None
    return sel_atoms.center_of_geometry()


# //////////////////////////////////////////////////////////////////////////////
class Triplet:
    def __init__(self,
        res: Residue, interactor: str,
        tail: tuple[str], head: str, hbond_fixed: bool
    ):
        self._tail = tail
        self._head = head
        self.interactor = interactor
        self.hbond_fixed = hbond_fixed

        self.pos_tail: np.ndarray | None = None
        self.pos_head: np.ndarray | None = None
        self.pos_interactor: np.ndarray | None = None

        self.resname = res.resname.upper()
        self.str_prev_res = f"segid {res.segid} and resid {res.resid - 1}"
        self.str_this_res = f"segid {res.segid} and resid {res.resid    }"
        self.str_next_res = f"segid {res.segid} and resid {res.resid + 1}"

    # --------------------------------------------------------------------------
    def set_pos_tail(self, atoms: mda.AtomGroup) -> np.ndarray | None:
        self.pos_tail = _safe_return_coords(
            atoms, f"name {' '.join(self._tail)}"
        )


    # --------------------------------------------------------------------------
    def set_pos_head(self, atoms: mda.AtomGroup) -> np.ndarray | None:
        self.pos_head = _safe_return_coords(
            atoms, f"name {self._head}"
        )


    # --------------------------------------------------------------------------
    def set_pos_interactor(self, atoms: mda.AtomGroup) -> np.ndarray | None:
        self.pos_interactor = _safe_return_coords(
            atoms, f"name {self.interactor}"
        )


    # --------------------------------------------------------------------------
    def set_pos_tail_custom(self,
        atoms: mda.AtomGroup, query_t0: str, query_t1: str
    ) -> np.ndarray | None:
        """
        Reserved for special cases in protein and RNA.
        """
        assert len(self._tail) == 2, "Custom tail position can only be set for triplets with two tail points."
        t0, t1 = self._tail
        self.pos_tail = _safe_return_coords(
            atoms,
            f"({query_t0} and name {t0}) or " +\
            f"({query_t1} and name {t1})"
        )


    # ------------------------------------------------------------------------------
    def get_interactor_bonded_hydrogens(self, atoms: mda.AtomGroup) -> tuple:
        sel_atoms = atoms.select_atoms(f"name {self.interactor}")
        if len(sel_atoms) == 0:
            return []
        bonded_atoms = [
            (bond.atoms[0] if bond.atoms[0].name != self.interactor else bond.atoms[1])
            for bond in sel_atoms.bonds
        ]
        return tuple(filter(lambda a: a.type == 'H', bonded_atoms))


    # --------------------------------------------------------------------------
    def get_direction_vector(self):
        if (self.pos_tail is None) or (self.pos_head is None):
            return None
        return vg.Math.normalize(self.pos_head - self.pos_tail)


# //////////////////////////////////////////////////////////////////////////////



import volgrids as vg
import volgrids.smiffer as sm

from .hydro import SmifHydro

# //////////////////////////////////////////////////////////////////////////////
class SmifHydrophilic(SmifHydro):
    def populate_grid(self):
        radius = sm.MU_HYDROPHILIC + sm.GAUSSIAN_KERNEL_SIGMAS * sm.SIGMA_HYDROPHILIC
        kernel = vg.KernelGaussianUnivariateDist(
            radius, self.ms.deltas, vg.FLOAT_DTYPE, sm.PARAMS_HPHIL
        )
        kernel.link_to_grid(self.grid, self.ms.minCoords)

        for particle, mul_factor in self.iter_particles():
            if mul_factor > 0: continue
            kernel.stamp(particle.position, multiplication_factor = -mul_factor)


# //////////////////////////////////////////////////////////////////////////////



from setuptools import setup, find_packages
from pathlib import Path

def read_requirements(path = "requirements.txt"):
    p = Path(path)
    if not p.exists(): return []
    stripped_lines = (ln.strip() for ln in p.read_text().strip().splitlines())
    return [ln for ln in stripped_lines if ln and not ln.startswith("#")]

setup(
    name="volgrids",
    version="0.4.1",
    description="Framework for volumetric calculations, with emphasis in biological molecular systems.",
    keywords="grid mif smif volumetric molecular structural biology interaction field",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="DiegoBarMor",
    author_email="diegobarmor42@gmail.com",
    url="https://github.com/diegobarmor/volgrids",
    license="MIT",
    packages=find_packages(),
    package_data={"volgrids": ["_tables/*", "utils/*"]},
    install_requires=read_requirements(),
    entry_points={
        "console_scripts": [
            "volgrids=volgrids.__main__:main",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.10",
)



import numpy as np

import volgrids as vg

# //////////////////////////////////////////////////////////////////////////////
class KernelSphere(vg.Kernel):
    """For generating simple boolean spheres (e.g. for masks)"""
    def __init__(self, radius, deltas, dtype):
        super().__init__(radius, deltas, dtype)
        self.kernel[self.dist < radius] = 1


# //////////////////////////////////////////////////////////////////////////////
class KernelCylinder(vg.Kernel):
    """For generating boolean cylinders"""
    def __init__(self, radius, vdirection, width, deltas, dtype):
        super().__init__(radius, deltas, dtype)
        w = vg.Math.get_projection_height(self.shifted_coords, vdirection)
        self.kernel[w < width] = 1


# //////////////////////////////////////////////////////////////////////////////
class KernelDisk(KernelSphere):
    """For generating boolean disks"""
    def __init__(self, radius, vnormal, height, deltas, dtype):
        super().__init__(radius, deltas, dtype)
        projection = vg.Math.get_projection(self.shifted_coords, vnormal)
        projection = np.abs(projection)
        self.kernel[projection >= height] = 0


# //////////////////////////////////////////////////////////////////////////////
class KernelDiskConecut(KernelDisk):
    """For generating boolean disks with a cone cut"""
    def __init__(self, radius, vnormal, height, vdirection, max_angle, deltas, dtype):
        super().__init__(radius, vnormal, height, deltas, dtype)
        angle = vg.Math.get_angle(self.shifted_coords, vdirection, in_degrees = False)
        self.kernel[angle > max_angle/2] = 0


# //////////////////////////////////////////////////////////////////////////////



from ._core.grid import Grid
from ._core.mol_system import MolSystem

from ._kernels.kernel import Kernel
from ._kernels.boolean import \
    KernelSphere, KernelCylinder, KernelDisk, KernelDiskConecut
from ._kernels.gaussian import \
    KernelGaussianUnivariateDist, KernelGaussianBivariateAngleDist

from ._misc.math import Math
from ._misc.params_gaussian import ParamsGaussian, \
    ParamsGaussianUnivariate, ParamsGaussianBivariate
from ._misc.timer import Timer
from ._misc.utils import resolve_path_package

from ._parsers.parser_ini import ParserIni
from ._parsers.parser_config import ParserConfig
from ._parsers.grid_io import GridFormat, GridIO

from ._ui.param_handler import ParamHandler
from ._ui.app import App

from ._misc.apbs_subprocess import APBSSubprocess


############################# CONFIG FILE GLOBALS ##############################
import numpy as _np
_keys_other = set(globals().keys())

OUTPUT_FORMAT: GridFormat = GridFormat.CMAP_PACKED

GZIP_COMPRESSION: int = 9
FLOAT_DTYPE: type = _np.float32
WARNING_GRID_SIZE: float = 5.0e7

GRID_DX: float = 0.25
GRID_DY: float = 0.25
GRID_DZ: float = 0.25

GRID_XRES: int = 200
GRID_YRES: int = 200
GRID_ZRES: int = 200

EXTRA_BOX_SIZE: int = 5
USE_FIXED_DELTAS: bool = True

__config_keys__ = set(globals().keys()) - _keys_other


######################## COMMAND LINE ARGUMENTS GLOBALS ########################
### These are global variables that are to be set by reading config files
### DEFAULT config.ini allows to first read "config_volgrids.ini" from the volgrid's repo root,
### to be used by the volgrid's main scripts. Its default remains None for any other use case.
### CUSTOM config.ini allows the user to specify a custom config file path from the command line.

import pathlib as _pathlib
PATH_DEFAULT_CONFIG:      _pathlib.Path  = None # "./config_volgrids.ini"
PATHS_CUSTOM_CONFIG: list[_pathlib.Path] = []   # "path/input/config.ini"


############################### RUNTIME GLOBALS ################################
PQR_CONTENTS_TEMP: str = ""
